import os

import torch
import random
import numpy as np
import scanpy as sc
import scipy.sparse as sp
from torch.backends import cudnn
#from scipy.sparse import issparse
from scipy.sparse.csc import csc_matrix
from scipy.sparse.csr import csr_matrix
from sklearn.neighbors import NearestNeighbors 

def filter_with_overlap_gene(adata, adata_sc):
    # remove all-zero-valued genes
    #sc.pp.filter_genes(adata, min_cells=1)
    #sc.pp.filter_genes(adata_sc, min_cells=1)
    
    if 'space_variable_features' in adata.var.keys():
        adata = adata[:, adata.var['space_variable_features']]
    elif 'highly_variable' in adata.var.keys():
        adata = adata[:, adata.var['highly_variable']]
    else:
        raise ValueError("'highly_variable' or 'space_variable_features' are not existed in adata!")

    if 'space_variable_features' in adata_sc.var.keys():
        adata_sc = adata_sc[:, adata_sc.var['space_variable_features']]
    elif 'highly_variable' in adata_sc.var.keys():
        adata_sc = adata_sc[:, adata_sc.var['highly_variable']]
    else:
        raise ValueError("'highly_variable' or 'space_variable_features' are not existed in adata_sc!")


    # Refine `marker_genes` so that they are shared by both adatas
    genes = list(set(adata.var.index) & set(adata_sc.var.index))
    genes.sort()
    print('Number of overlap genes:', len(genes))

    adata.uns["overlap_genes"] = genes
    adata_sc.uns["overlap_genes"] = genes
    
    adata = adata[:, genes]
    adata_sc = adata_sc[:, genes]
    
    return adata, adata_sc

def permutation(feature):
    # fix_seed(FLAGS.random_seed) 
    ids = np.arange(feature.shape[0])
    ids = np.random.permutation(ids)
    feature_permutated = feature[ids]
    
    return feature_permutated 

def construct_interaction(adata, n_neighbors=3):
    """Constructing spot-to-spot interactive graph"""
    import ot
    position = adata.obsm['spatial']
    
    # calculate distance matrix
    distance_matrix = ot.dist(position, position, metric='euclidean')
    n_spot = distance_matrix.shape[0]
    
    adata.obsm['distance_matrix'] = distance_matrix
    
    # find k-nearest neighbors
    interaction = np.zeros([n_spot, n_spot])  
    for i in range(n_spot):
        vec = distance_matrix[i, :]
        distance = vec.argsort()
        for t in range(1, n_neighbors + 1):
            y = distance[t]
            interaction[i, y] = 1
         
    adata.obsm['graph_neigh'] = interaction
    
    #transform adj to symmetrical adj
    adj = interaction
    adj = adj + adj.T
    adj = np.where(adj>1, 1, adj)
    
    adata.obsm['adj'] = adj
    
def construct_interaction_KNN(adata, n_neighbors=3):
    position = adata.obsm['spatial']
    n_spot = position.shape[0]
    nbrs = NearestNeighbors(n_neighbors=n_neighbors+1).fit(position)  
    _ , indices = nbrs.kneighbors(position)
    x = indices[:, 0].repeat(n_neighbors)
    y = indices[:, 1:].flatten()
    interaction = np.zeros([n_spot, n_spot])
    interaction[x, y] = 1
    
    adata.obsm['graph_neigh'] = interaction
    
    #transform adj to symmetrical adj
    adj = interaction
    adj = adj + adj.T
    adj = np.where(adj>1, 1, adj)
    
    adata.obsm['adj'] = adj
    print('Graph constructed!')   

def preprocess(adata):
    sc.pp.highly_variable_genes(adata, flavor="seurat_v3", n_top_genes=3000)
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.scale(adata, zero_center=False, max_value=10)
    
def get_feature(adata, deconvolution=False):
    if deconvolution:
        adata_Vars = adata
    else:   

        if 'space_variable_features' in adata.var.keys():
            adata_Vars = adata[:, adata.var['space_variable_features']]
        elif 'highly_variable' in adata.var.keys():
            adata_Vars = adata[:, adata.var['highly_variable']]
        else:
            raise ValueError("'highly_variable' or 'space_variable_features' are not existed in adata!")
        
       
    if isinstance(adata_Vars.X, csc_matrix) or isinstance(adata_Vars.X, csr_matrix):
       feat = adata_Vars.X.toarray()[:, ]
    else:
       feat = adata_Vars.X[:, ] 
    
    # data augmentation
    feat_a = permutation(feat)
    
    adata.obsm['feat'] = feat
    adata.obsm['feat_a'] = feat_a    
    
def add_contrastive_label(adata):
    # contrastive label
    n_spot = adata.n_obs
    one_matrix = np.ones([n_spot, 1])
    zero_matrix = np.zeros([n_spot, 1])
    label_CSL = np.concatenate([one_matrix, zero_matrix], axis=1)
    adata.obsm['label_CSL'] = label_CSL
    
def normalize_adj(adj):
    """Symmetrically normalize adjacency matrix."""
    adj = sp.coo_matrix(adj)
    rowsum = np.array(adj.sum(1))
    d_inv_sqrt = np.power(rowsum, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
    d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
    adj = adj.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt)
    return adj.toarray()

def preprocess_adj(adj):
    """Preprocessing of adjacency matrix for simple GCN model and conversion to tuple representation."""
    adj_normalized = normalize_adj(adj)+np.eye(adj.shape[0])
    return adj_normalized 

def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)

def preprocess_adj_sparse(adj):
    adj = sp.coo_matrix(adj)
    adj_ = adj + sp.eye(adj.shape[0])
    rowsum = np.array(adj_.sum(1))
    degree_mat_inv_sqrt = sp.diags(np.power(rowsum, -0.5).flatten())
    adj_normalized = adj_.dot(degree_mat_inv_sqrt).transpose().dot(degree_mat_inv_sqrt).tocoo()
    return sparse_mx_to_torch_sparse_tensor(adj_normalized)    

def fix_seed(seed):
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    cudnn.deterministic = True
    cudnn.benchmark = False
    
    os.environ['PYTHONHASHSEED'] = str(seed)
    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8' 
    
    



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import scanpy as sc
from scipy.stats import mode
from sklearn.neighbors import KNeighborsClassifier

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class KNN_BBKNN(BaseAlgorithm):
    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        result_key: str | None = "popv_knn_bbknn_prediction",
        umap_key: str | None = "X_umap_bbknn_popv",
        method_kwargs: dict | None = None,
        classifier_kwargs: dict | None = None,
        embedding_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            umap_key=umap_key,
        )
        self.method_kwargs = {
            "metric": "euclidean",
            "approx": True,
            "n_pcs": 50,
            "neighbors_within_batch": 3,
            "use_annoy": False,
        }
        if method_kwargs:
            self.method_kwargs.update(method_kwargs)

        self.classifier_kwargs = {"weights": "uniform", "n_neighbors": 15}
        if classifier_kwargs:
            self.classifier_kwargs.update(classifier_kwargs)

        self.embedding_kwargs = {"min_dist": 0.1}
        if embedding_kwargs:
            self.embedding_kwargs.update(embedding_kwargs)

    def compute_integration(self, adata):
        """Compute BBKNN integration."""
        logging.info("Integrating data with bbknn")
        # Use the available umap key from the reference integration.
        ref_umap_key = "X_umap_bbknn"
        if (
            adata.uns.get("_prediction_mode") == "inference"
            and ref_umap_key in adata.obsm
            and not settings.recompute_embeddings
        ):
            # Cache masks for query and reference cells
            query_mask = adata.obs["_dataset"] == "query"
            ref_mask = adata.obs["_dataset"] == "ref"

            index_path = os.path.join(adata.uns["_save_path_trained_models"], "pynndescent_index.joblib")
            index = joblib.load(index_path)

            query_features = adata.obsm["X_pca"][query_mask].astype(np.float32)
            indices, _ = index.query(query_features, k=5)

            ref_umap = adata.obsm[ref_umap_key][ref_mask].astype(np.float32)
            neighbor_embedding = ref_umap[indices]
            # Compute mean embedding across neighbors and assign
            adata.obsm[self.umap_key][query_mask] = np.mean(neighbor_embedding, axis=1)
            adata.obsm[self.umap_key] = adata.obsm[self.umap_key].astype(np.float32)

            ref_probs = adata.obs[f"{self.result_key}_probabilities"][ref_mask].astype(np.float32)
            neighbor_probabilities = ref_probs[indices]
            adata.obs.loc[query_mask, f"{self.result_key}_probabilities"] = np.mean(
                neighbor_probabilities, axis=1
            )

            ref_prediction = adata.obs[f"{self.result_key}"][ref_mask]
            neighbor_prediction = ref_prediction.iloc[indices].to_numpy()
            mode_vals = mode(neighbor_prediction, axis=1).mode.ravel()
            adata.obs.loc[query_mask, f"{self.result_key}"] = mode_vals
        else:
            # If too many batches, reduce settings to avoid memory issues.
            batches = adata.obs[self.batch_key].unique()
            if len(batches) > 100:
                logging.warning("Using PyNNDescent instead of FAISS due to high number of batches.")
                self.method_kwargs["neighbors_within_batch"] = 1
                self.method_kwargs["pynndescent_n_neighbors"] = 10
                sc.external.pp.bbknn(
                    adata, batch_key=self.batch_key, use_faiss=False, use_rep="X_pca", **self.method_kwargs
                )
            else:
                sc.external.pp.bbknn(
                    adata, batch_key=self.batch_key, use_faiss=True, use_rep="X_pca", **self.method_kwargs
                )

    def predict(self, adata):
        """Predict celltypes using KNN based on BBKNN results."""
        logging.info(f'Saving knn on bbknn results to adata.obs["{self.result_key}"]')
        distances = adata.obsp["distances"]
        # Get indices of reference cells
        ref_indices = np.where(adata.obs["_labelled_train_indices"])[0]
        train_y = adata.obs.loc[adata.obs["_labelled_train_indices"], self.labels_key].cat.codes.to_numpy()

        train_distances = distances[ref_indices, :][:, ref_indices]
        test_distances = distances[:, :][:, ref_indices]

        # Adjust KNN neighbors if BBKNN returned a smaller graph
        train_counts = np.diff(train_distances.indptr).min()
        test_counts = np.diff(test_distances.indptr).min()
        smallest_graph = min(train_counts, test_counts)
        if smallest_graph < self.classifier_kwargs["n_neighbors"]:
            logging.warning(f"BBKNN found only {smallest_graph} neighbors. Reducing KNN n_neighbors.")
            self.classifier_kwargs["n_neighbors"] = smallest_graph

        knn = KNeighborsClassifier(metric="precomputed", **self.classifier_kwargs)
        knn.fit(train_distances, y=train_y)
        predicted_codes = knn.predict(test_distances)
        adata.obs[self.result_key] = adata.uns["label_categories"][predicted_codes]

        if self.return_probabilities:
            probabilities = knn.predict_proba(test_distances).max(axis=1)
            adata.obs[f"{self.result_key}_probabilities"] = probabilities

    def compute_umap(self, adata):
        """Compute UMAP embedding of integrated data."""
        if self.compute_umap_embedding:
            logging.info(f'Saving UMAP of bbknn results to adata.obsm["{self.umap_key}"]')
            if len(adata.obs[self.batch_key]) < 30 and settings.cuml:
                method = "rapids"
            else:
                logging.warning("Using UMAP instead of RAPIDS due to high number of batches.")
                method = "umap"
            # Compute UMAP embedding with a copy and assign the embedded coordinates directly.
            embedded = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs).obsm["X_umap"]
            adata.obsm[self.umap_key] = embedded



"""Collection of optional external algorithms."""
from . import (scSLAT,CEFCON,mofapy2,GNTD,spaceflow,STT,
               tosica,STAGATE_pyG,STAligner,spatrio,PROST,cytotrace2,
               GraphST,commot,cnmf,starfysh,scMulan,flowsig,PyWGCNA,
               CAST,scanorama,scdiffusion,BINARY,cellanova,VIA,gaston
               )
__all__ = [
    'scSLAT',
    'CEFCON',
    'mofapy2',
    'GNTD',
    'spaceflow',
    'STT',
    'tosica',
    'STAGATE_pyG',
    'STAligner',
    'spatrio',
    'PROST',
    'cytotrace2',
    'GraphST',
    'commot',
    'cnmf',
    'starfysh',
    'scMulan',
    'flowsig',
    'PyWGCNA',
    'CAST',
    'scanorama',
    'scdiffusion',
    'BINARY',
    'cellanova',
    'VIA',
    'gaston',
]



from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy
import os
import sys

# 查找是否存在.pyx文件而不是.py文件
force_dir = os.path.dirname(os.path.abspath(__file__))
has_pyx = os.path.exists(os.path.join(force_dir, "fa2util.pyx"))
source_file = "fa2util.pyx" if has_pyx else "fa2util.py"

# 定义编译参数
compiler_directives = {
    'language_level': 3,
    'boundscheck': False,  # 关闭数组边界检查 (提高性能)
    'wraparound': False,   # 关闭负索引自动处理 (提高性能)
    'initializedcheck': False,  # 关闭初始化检查 (提高性能)
    'cdivision': True,     # 使用C除法而不是Python除法 (提高性能)
    'fastcall': True       # 优化函数调用 (提高性能)
}

# 定义额外的编译参数
extra_compile_args = []
if sys.platform == 'win32':
    extra_compile_args.append('/O2')  # Windows: 启用优化
else:
    extra_compile_args.extend(['-O3', '-ffast-math'])  # Unix: 启用高级优化

# 定义扩展模块
extensions = [
    Extension(
        "fa2util",  # 不要包含.pyx后缀
        [source_file],  # Cython源文件
        include_dirs=[numpy.get_include()],  # 添加numpy头文件
        language="c++",  # 使用C++编译
        extra_compile_args=extra_compile_args
    )
]

print(f"正在编译 {source_file}...")
print(f"编译器指令: {compiler_directives}")
print(f"额外编译参数: {extra_compile_args}")

# 安装
setup(
    name="forceatlas2",
    version="1.0.0",
    description="高性能ForceAtlas2算法实现",
    author="OMICVerse Team",
    ext_modules=cythonize(
        extensions,
        compiler_directives=compiler_directives,
        annotate=False  # 设置为True可以生成html注释文件
    ),
    zip_safe=False
) 


import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from sklearn import linear_model
import networkx as nx
import seaborn as sns
import pandas as pd
from sklearn import decomposition as skldec 

from ..utils import plot_text_set
import matplotlib

def geneset_enrichment(gene_list:list,pathways_dict:dict,
                       pvalue_threshold:float=0.05,pvalue_type:str='auto',
                       organism:str='Human',description:str='None',
                       background:list=None,
                       outdir:str='./enrichr',cutoff:float=0.5)->pd.DataFrame:
    """
    Performs gene set enrichment analysis using Enrichr API.

    Arguments:
        gene_list: List of gene symbols to be tested for enrichment.
        pathways_dict: Dictionary of pathway library names and corresponding Enrichr API URLs.
        pvalue_threshold: P-value threshold for significant pathways. Default is 0.05.
        pvalue_type: Type of p-value correction to use. 'auto' uses Benjamini-Hochberg correction,for small gene sets (<500 genes) and Bonferroni correction for larger gene sets.,'bh' uses only Benjamini-Hochberg correction. 'bonferroni' uses only Bonferroni correction.,Default is 'auto'.
        organism: Organism of the input gene list. Default is 'Human'.
        description: Description of the input gene list. Default is 'None'.
        background: Background gene list to use for enrichment analysis. Default is None. If None, the background gene list is automatically set to the organism-specific gene list.
        outdir: Output directory for Enrichr results. Default is './enrichr'.
        cutoff: Show enriched terms which Adjusted P-value < cutoff. Default is 0.5.

    Returns:
        enrich_res: A pandas DataFrame containing the enrichment results.


    """
    from ..externel.gseapy import enrichr
    #import gseapy as gp
    if background is None:
        if (organism == 'Mouse') or (organism == 'mouse') or (organism == 'mm'):
            background='mmusculus_gene_ensembl'
        elif (organism == 'Human') or (organism == 'human') or (organism == 'hs'):
            background='hsapiens_gene_ensembl'

    enr = enrichr(gene_list=gene_list,
                 gene_sets=pathways_dict,
                 organism=organism, # don't forget to set organism to the one you desired! e.g. Yeast
                 description=description,
                 background=background,
                 outdir=outdir,
                 cutoff=cutoff # test dataset, use lower value from range(0,1)
                )
    if pvalue_type=='auto':
        if enr.res2d.shape[0]>100:
            enrich_res=enr.res2d[enr.res2d['Adjusted P-value']<pvalue_threshold]
            enrich_res['logp']=-np.log(enrich_res['Adjusted P-value'])
        else:
            enrich_res=enr.res2d[enr.res2d['P-value']<pvalue_threshold]
            enrich_res['logp']=-np.log(enrich_res['P-value'])
    elif pvalue_type=='adjust':
        enrich_res=enr.res2d[enr.res2d['Adjusted P-value']<pvalue_threshold]
        enrich_res['logp']=-np.log(enrich_res['Adjusted P-value'])
    else:
        enrich_res=enr.res2d[enr.res2d['P-value']<pvalue_threshold]
        enrich_res['logp']=-np.log(enrich_res['P-value'])
    enrich_res['logc']=np.log(enrich_res['Odds Ratio'])
    enrich_res['num']=[int(i.split('/')[0]) for i in enrich_res['Overlap']]
    enrich_res['fraction']=[int(i.split('/')[0])/int(i.split('/')[1]) for i in enrich_res['Overlap']]
    return enrich_res

def enrichment_multi_concat(enr_dict):
    def process_df(df, term_col_name):
        new_data = []
        for _, row in df.iterrows():
            genes = row['Genes'].split(';')
            for gene in genes:
                new_data.append({
                    'Gene': gene,
                    term_col_name: row['Term'],
                    
                })
        return pd.DataFrame(new_data)
    new_dict={}
    new_li=[]
    for key in enr_dict.keys():
        new_dict[key]=process_df(enr_dict[key], key)
        new_li.append(new_dict[key])
    # 合并两个DataFrame
    merged_df = pd.concat(new_li, ignore_index=True).fillna('')
    #return merged_df

    #print(dict(zip(enr_dict.keys(),
    #        [lambda x: '|'.join(x.dropna().unique()) for i in range(len(enr_dict.keys()))])))
    
    # 按基因分组，并将相同基因的Term合并
    result_df = merged_df.groupby('Gene').agg(dict(zip(enr_dict.keys(),
            [lambda x: '|'.join(x.dropna().unique()) for i in range(len(enr_dict.keys()))]))
                                             ).reset_index()
    return result_df



def geneset_enrichment_GSEA(gene_rnk:pd.DataFrame,pathways_dict:dict,
                            processes:int=8,
                     permutation_num:int=100, # reduce number to speed up testing
                     outdir:str='./enrichr_gsea', format:str='png', seed:int=112)->dict:
    """
    Enrichment analysis using GSEA

    Arguments:
        gene_rnk: pre-ranked correlation table or pandas DataFrame. Same input with ``GSEA`` .rnk file.
        pathways_dict: Dictionary of pathway library names and corresponding Enrichr API URLs.
        processes: Number of Processes you are going to use. Default: 8.
        permutation_num: Number of permutations for significance computation. Default: 100.
        outdir: Output directory for Enrichr results. Default is './enrichr_gsea'.
        format: Matplotlib figure format. Default: 'png'.
        seed: Random seed. Default: 112.

    Returns:
        pre_res: A prerank object containing the enrichment results.
    
    """
    from ..externel.gseapy import prerank
    pre_res = prerank(rnk=gene_rnk, gene_sets=pathways_dict,
                     processes=processes,
                     permutation_num=permutation_num, # reduce number to speed up testing
                     outdir=outdir, format=format, seed=seed)
    return pre_res
    enrich_res=pre_res.res2d[pre_res.res2d['fdr']<0.05]
    enrich_res['logp']=-np.log(enrich_res['fdr']+0.0001)
    enrich_res['logc']=enrich_res['nes']
    enrich_res['num']=enrich_res['matched_size']
    enrich_res['fraction']=enrich_res['matched_size']/enrich_res['geneset_size']
    enrich_res['Term']=enrich_res.index.tolist()
    enrich_res['P-value']=enrich_res['fdr']
    return enrich_res

def geneset_plot_multi(enr_dict,colors_dict,num:int=5,fontsize=10,
                        fig_title:str='',fig_xlabel:str='Fractions of genes',
                        figsize:tuple=(2,4),cmap:str='YlGnBu',
                        text_knock:int=5,text_maxsize:int=20,ax=None,
                        ):
    """
    Enrichment multi genesets analysis using GSEA

    Arguments:
        enr_dict: A dictionary of enrichment results.
        colors_dict: A dictionary of colors for each gene set.
        num: The number of enriched terms to plot. Default is 5.
        fontsize: The fontsize of the plot. Default is 10.
        fig_title: The title of the plot. Default is an empty string.
        fig_xlabel: The label of the x-axis. Default is 'Fractions of genes'.
        figsize: The size of the plot. Default is (2,4).
        cmap: The colormap to use for the plot. Default is 'YlGnBu'.
        text_knock: The number of characters to knock off the end of the term name. Default is 5.
        text_maxsize: The maximum fontsize of the term names. Default is 20.
        ax: A matplotlib.axes.Axes object.

    """
    from PyComplexHeatmap import HeatmapAnnotation,DotClustermapPlotter,anno_label,anno_simple,AnnotationBase
    for key in enr_dict.keys():
        enr_dict[key]['Type']=key
    enr_all=pd.concat([enr_dict[i].iloc[:num] for i in enr_dict.keys()],axis=0)
    enr_all['Term']=[plot_text_set(i.split('(')[0],text_knock=text_knock,text_maxsize=text_maxsize) for i in enr_all.Term.tolist()]
    enr_all.index=enr_all.Term
    enr_all['Term1']=[i for i in enr_all.index.tolist()]
    del enr_all['Term']

    colors=colors_dict

    left_ha = HeatmapAnnotation(
                          label=anno_label(enr_all.Type, merge=True,rotation=0,colors=colors,relpos=(1,0.8)),
                          Category=anno_simple(enr_all.Type,cmap='Set1',
                                           add_text=False,legend=False,colors=colors),
                           axis=0,verbose=0,label_kws={'rotation':45,'horizontalalignment':'left','visible':False})
    right_ha = HeatmapAnnotation(
                              label=anno_label(enr_all.Term1, merge=True,rotation=0,relpos=(0,0.5),arrowprops=dict(visible=True),
                                               colors=enr_all.assign(color=enr_all.Type.map(colors)).set_index('Term1').color.to_dict(),
                                              fontsize=fontsize,luminance=0.8,height=2),
                               axis=0,verbose=0,#label_kws={'rotation':45,'horizontalalignment':'left'},
                                orientation='right')
    if ax==None:
        fig, ax = plt.subplots(figsize=figsize) 
    else:
        ax=ax
    #plt.figure(figsize=figsize)
    cm = DotClustermapPlotter(data=enr_all, x='fraction',y='Term1',value='logp',c='logp',s='num',
                              cmap=cmap,
                              row_cluster=True,#col_cluster=True,#hue='Group',
                              #cmap={'Group1':'Greens','Group2':'OrRd'},
                              vmin=-1*np.log10(0.1),vmax=-1*np.log10(1e-10),
                              #colors={'Group1':'yellowgreen','Group2':'orange'},
                              #marker={'Group1':'*','Group2':'$\\ast$'},
                              show_rownames=True,show_colnames=False,row_dendrogram=False,
                              col_names_side='top',row_names_side='right',
                              xticklabels_kws={'labelrotation': 30, 'labelcolor': 'blue','labelsize':fontsize},
                              #yticklabels_kws={'labelsize':10},
                              #top_annotation=col_ha,left_annotation=left_ha,right_annotation=right_ha,
                              left_annotation=left_ha,right_annotation=right_ha,
                              spines=False,
                              row_split=enr_all.Type,# row_split_gap=1,
                              #col_split=df_col.Group,col_split_gap=0.5,
                              verbose=1,legend_gap=10,
                              #dot_legend_marker='*',
                              xlabel='Fractions of genes',xlabel_side="bottom",
                              xlabel_kws=dict(labelpad=8,fontweight='normal',fontsize=fontsize+2),
                              # xlabel_bbox_kws=dict(facecolor=facecolor)
                             )
    tesr=plt.gcf().axes
    for ax in plt.gcf().axes:
        if hasattr(ax, 'get_xlabel'):
            if ax.get_xlabel() == 'Fractions of genes':  # 假设 colorbar 有一个特定的标签
                cbar = ax
                cbar.grid(False)
            if ax.get_ylabel() == 'logp':  # 假设 colorbar 有一个特定的标签
                cbar = ax
                cbar.tick_params(labelsize=fontsize+2)
                cbar.set_ylabel(r'$−Log_{10}(P_{adjusted})$',fontsize=fontsize+2)
                cbar.grid(False)
    return ax

def geneset_plot(enrich_res,num:int=10,node_size:list=[5,10,15],
                        cax_loc:list=[2, 0.55, 0.5, 0.02],cax_fontsize:int=12,
                        fig_title:str='',fig_xlabel:str='Fractions of genes',
                        figsize:tuple=(2,4),cmap:str='YlGnBu',
                        text_knock:int=5,text_maxsize:int=20,
                        bbox_to_anchor_used:tuple=(-0.45, -13),node_diameter:int=10,
                        custom_ticks:list=[5,10],ax=None)->matplotlib.axes._axes.Axes:
    """
    Plot the gene set enrichment result.

    Arguments:
        num: The number of enriched terms to plot. Default is 10.
        node_size: A list of integers defining the size of nodes in the plot. Default is [5,10,15].
        cax_loc: The location, width and height of the colorbar on the plot. Default is [2, 0.55, 0.5, 0.02].
        cax_fontsize: The fontsize of the colorbar label. Default is 12.
        fig_title: The title of the plot. Default is an empty string.
        fig_xlabel: The label of the x-axis. Default is 'Fractions of genes'.
        figsize: The size of the plot. Default is (2,4).
        cmap: The colormap to use for the plot. Default is 'YlGnBu'.
        text_knock: The number of characters to knock off the end of the term name. Default is 2.
        text_maxsize: The maximum fontsize of the term names. Default is 20.
        bbox_to_anchor_used (tuple): The anchor point for placing the legend. Default is (-0.45, -13).
        node_diameter (int): The base size for nodes in the plot. Default is 10.

    Returns:
        A matplotlib.axes.Axes object.
    
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    plot_data2=enrich_res.sort_values('P-value')[:num].sort_values('logc')
    st=ax.scatter(plot_data2['fraction'],range(len(plot_data2['logc'])),
            s=plot_data2['num']*node_diameter,linewidths=1,edgecolors='black',c=plot_data2['logp'],cmap=cmap)
    ax.yaxis.tick_right()
    plt.yticks(range(len(plot_data2['fraction'])),[plot_text_set(i.split('(')[0],text_knock=text_knock,text_maxsize=text_maxsize) for i in plot_data2['Term']],
            fontsize=10,)
    plt.xticks(fontsize=12,)
    plt.title(fig_title,fontsize=12)
    plt.xlabel(fig_xlabel,fontsize=12)

    fig = plt.gcf()
    cax = fig.add_axes(cax_loc)
    cb=fig.colorbar(st,shrink=0.25,cax=cax,orientation='horizontal')
    cb.set_label(r'$−Log_{10}(P_{adjusted})$',fontdict={'size':cax_fontsize})
    # new code to add custom ticks
    cb.set_ticks(custom_ticks)

    gl_li=[]
    for i in node_size:
        gl_li.append(ax.scatter([],[], s=i*node_diameter, marker='o', color='white',edgecolors='black'))

    plt.legend(gl_li,
        [str(i) for i in node_size],
        loc='lower left',
        ncol=3,bbox_to_anchor=bbox_to_anchor_used,
        fontsize=cax_fontsize)
    return ax

class pyGSE(object):

    def __init__(self,gene_list:list,pathways_dict:dict,pvalue_threshold:float=0.05,pvalue_type:str='auto',
                 background=None,organism:str='Human',description:str='None',outdir:str='./enrichr',cutoff:float=0.5) -> None:
        """Initialize the pyGSE class.

        Arguments:
            gene_list: A list of genes.
            pathways_dict: A dictionary of pathways.
            pvalue_threshold: The p-value threshold for enrichment. Default is 0.05.
            pvalue_type: The p-value type. Default is 'auto'.
            organism: The organism. Default is 'Human'.
            description: The description. Default is 'None'.
            outdir: The output directory. Default is './enrichr'.
            cutoff: The cutoff for enrichment. Default is 0.5.
        
        """

        self.gene_list=gene_list
        self.pathways_dict=pathways_dict
        self.pvalue_threshold=pvalue_threshold
        self.pvalue_type=pvalue_type
        self.organism=organism
        self.description=description
        self.outdir=outdir
        self.cutoff=cutoff
        if background is None:
            if (organism == 'Mouse') or (organism == 'mouse') or (organism == 'mm'):
                background='mmusculus_gene_ensembl'
            elif (organism == 'Human') or (organism == 'human') or (organism == 'hs'):
                background='hsapiens_gene_ensembl'
            self.background=background
        else:
            self.background=background
    
    def enrichment(self):
        """gene set enrichment analysis.
        
        Returns:
            A pandas.DataFrame object containing the enrichment results.
        """

        enrich_res=geneset_enrichment(self.gene_list,self.pathways_dict,self.pvalue_threshold,self.pvalue_type,
                                  self.organism,self.description,self.background,self.outdir,self.cutoff)
        self.enrich_res=enrich_res
        return enrich_res
    
    
    def plot_enrichment(self,num:int=10,node_size:list=[5,10,15],
                        cax_loc:int=2,cax_fontsize:int=12,
                        fig_title:str='',fig_xlabel:str='Fractions of genes',
                        figsize:tuple=(2,4),cmap:str='YlGnBu',text_knock:int=2,text_maxsize:int=20)->matplotlib.axes._axes.Axes:
        
        """Plot the gene set enrichment result.
        
        Arguments:
            num: The number of enriched terms to plot. Default is 10.
            node_size: A list of integers defining the size of nodes in the plot. Default is [5,10,15].
            cax_loc: The location of the colorbar on the plot. Default is 2.
            cax_fontsize: The fontsize of the colorbar label. Default is 12.
            fig_title: The title of the plot. Default is an empty string.
            fig_xlabel: The label of the x-axis. Default is 'Fractions of genes'.
            figsize: The size of the plot. Default is (2,4).
            cmap: The colormap to use for the plot. Default is 'YlGnBu'.

        Returns:
            A matplotlib.axes.Axes object.
        """
        return geneset_plot(self.enrich_res,num,node_size,cax_loc,cax_fontsize,
                            fig_title,fig_xlabel,figsize,cmap,text_knock,text_maxsize)
    
class pyGSEA(object):

    def __init__(self,gene_rnk:pd.DataFrame,pathways_dict:dict,
                 processes:int=8,permutation_num:int=100,
                 outdir:str='./enrichr_gsea',cutoff:float=0.5) -> None:
        """Initialize the pyGSEA class.

        Arguments:
            gene_rnk: pre-ranked correlation table or pandas DataFrame. Same input with ``GSEA`` .rnk file.
            pathways_dict: Dictionary of pathway library names and corresponding Enrichr API URLs.
            processes: Number of Processes you are going to use. Default: 8.
            permutation_num: Number of permutations for significance computation. Default: 100.
            outdir: Output directory for Enrichr results. Default is './enrichr_gsea'.
            cutoff: The cutoff for enrichment. Default is 0.5.
        """

        self.gene_rnk=gene_rnk
        self.pathways_dict=pathways_dict
        self.processes=processes
        self.permutation_num=permutation_num
        self.outdir=outdir
        self.cutoff=cutoff
    
    
    def enrichment(self,format:str='png', pval=0.05,seed:int=112)->pd.DataFrame:
        """gene set enrichment analysis.
        
        Arguments:
            format: Matplotlib figure format. Default: 'png'.
            seed: Random seed. Default: 112.
        
        Returns:
            enrich_res:A pandas.DataFrame object containing the enrichment results.
        """

        
        pre_res=geneset_enrichment_GSEA(self.gene_rnk,self.pathways_dict,
                                           self.processes,self.permutation_num,
                                           self.outdir,format,seed)
        self.pre_res=pre_res
        enrich_res=pre_res.res2d[pre_res.res2d['fdr']<pval]
        enrich_res['logp']=-np.log(enrich_res['fdr']+0.0001)
        enrich_res['logc']=enrich_res['nes']
        enrich_res['num']=enrich_res['matched_size']
        enrich_res['fraction']=enrich_res['matched_size']/enrich_res['geneset_size']
        enrich_res['Term']=enrich_res.index.tolist()
        enrich_res['P-value']=enrich_res['fdr']
        self.enrich_res=enrich_res
        return enrich_res
    
    def plot_gsea(self,term_num:int=0,
                  gene_set_title:str='',
                  figsize:tuple=(3,4),
                  cmap:str='RdBu_r',
                  title_fontsize:int=12,
                  title_y:float=0.95)->matplotlib.figure.Figure:
        """Plot the gene set enrichment result.
        
        Arguments:
            term_num: The number of enriched terms to plot. Default is 0.
            gene_set_title: The title of the plot. Default is an empty string.
            figsize: The size of the plot. Default is (3,4).
            cmap: The colormap to use for the plot. Default is 'RdBu_r'.
            title_fontsize: The fontsize of the title. Default is 12.
            title_y: The y coordinate of the title. Default is 0.95.

        Returns:
            fig: A matplotlib.figure.Figure object.
        """
        from ..externel.gseapy.plot import GSEAPlot
        #from gseapy.plot import GSEAPlot
        terms = self.enrich_res.index
        g = GSEAPlot(
        rank_metric=self.pre_res.ranking, term=terms[term_num],figsize=figsize,cmap=cmap,
            **self.pre_res.results[terms[term_num]]
            )
        if gene_set_title=='':
            g.fig.suptitle(terms[term_num],fontsize=title_fontsize,y=title_y)
        else:
            g.fig.suptitle(gene_set_title,fontsize=title_fontsize,y=title_y)
        g.add_axes()
        return g.fig
    
    
    def plot_enrichment(self,num:int=10,node_size:list=[5,10,15],
                        cax_loc:int=2,cax_fontsize:int=12,
                        fig_title:str='',fig_xlabel:str='Fractions of genes',
                        figsize:tuple=(2,4),cmap:str='YlGnBu',
                        text_knock:int=2,text_maxsize:int=20)->matplotlib.axes._axes.Axes:
        
        """Plot the gene set enrichment result.
        
        Arguments:
            num: The number of enriched terms to plot. Default is 10.
            node_size: A list of integers defining the size of nodes in the plot. Default is [5,10,15].
            cax_loc: The location of the colorbar on the plot. Default is 2.
            cax_fontsize: The fontsize of the colorbar label. Default is 12.
            fig_title: The title of the plot. Default is an empty string.
            fig_xlabel: The label of the x-axis. Default is 'Fractions of genes'.
            figsize: The size of the plot. Default is (2,4).
            cmap: The colormap to use for the plot. Default is 'YlGnBu'.
            text_knock: The number of terms to knock out for text labels. Default is 2.
            text_maxsize: The maximum fontsize of text labels. Default is 20.

        Returns:
            ax: A matplotlib.axes.Axes object.
        """
        return geneset_plot(self.enrich_res,num,node_size,cax_loc,cax_fontsize,
                            fig_title,fig_xlabel,figsize,cmap,text_knock,text_maxsize)


import torch,random
from tqdm import tqdm
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.sparse import csr_matrix
from sklearn.metrics import pairwise_distances,pairwise_distances_chunked,confusion_matrix
import scanpy as sc
from scipy.sparse import csr_matrix as csr
from .utils import coords2adjacentmat

def space_project(
    sdata_inte,
    idx_source,
    idx_target,
    raw_layer,
    source_sample,
    target_sample,
    coords_source,
    coords_target,
    output_path,
    source_sample_ctype_col,
    target_cell_pc_feature = None,
    source_cell_pc_feature = None,
    k2 = 1,
    ifplot = True,
    umap_feature = 'X_umap',
    ave_dist_fold = 2,
    batch_t = '',
    alignment_shift_adjustment = 50,
    color_dict = None,
    adjust_shift = False,
    metric_t = 'cosine',
    working_memory_t = 1000
    ):
    sdata_ref = sdata_inte[idx_target,:].copy()
    source_feat = sdata_inte[idx_source,:].layers[raw_layer].toarray()

    project_ind = np.zeros([np.sum(idx_target),k2]).astype(int)
    project_weight = np.zeros_like(project_ind).astype(float)
    cdists = np.zeros_like(project_ind).astype(float)
    physical_dist = np.zeros_like(project_ind).astype(float)
    all_avg_feat = np.zeros([np.sum(idx_target),source_feat.shape[1]]).astype(float)

    if source_sample_ctype_col is not None:
        for ctype_t in np.unique(sdata_inte[idx_target].obs[source_sample_ctype_col]):
            print(f'Start to project {ctype_t} cells:')
            idx_ctype_t = np.isin(sdata_inte[idx_target].obs[source_sample_ctype_col],ctype_t)
            ave_dist_t,_,_,_ = average_dist(coords_target[idx_ctype_t,:].copy(),working_memory_t=working_memory_t)
            dist_thres = ave_dist_fold * ave_dist_t + alignment_shift_adjustment
            if adjust_shift:
                coords_shift = group_shift(target_cell_pc_feature[idx_ctype_t,:], source_cell_pc_feature, coords_target[idx_ctype_t,:], coords_source, working_memory_t = working_memory_t, metric_t = metric_t)
                coords_source_t = coords_source + coords_shift
                print(coords_shift)
            else:
                coords_source_t = coords_source.copy()
            project_ind[idx_ctype_t,:],project_weight[idx_ctype_t,:],cdists[idx_ctype_t,:],physical_dist[idx_ctype_t,:],all_avg_feat[idx_ctype_t,:] = physical_dist_priority_project(
                feat_target = target_cell_pc_feature[idx_ctype_t,:],
                feat_source = source_cell_pc_feature,
                coords_target = coords_target[idx_ctype_t,:],
                coords_source = coords_source_t,
                source_feat = source_feat,
                k2 = 1,
                pdist_thres = dist_thres,
                metric_t = metric_t,
                working_memory_t = working_memory_t)
    else:
        ave_dist_t,_,_,_ = average_dist(coords_target.copy(),working_memory_t=working_memory_t,strategy_t='delaunay')
        dist_thres = ave_dist_fold * ave_dist_t + alignment_shift_adjustment
        project_ind,project_weight,cdists,physical_dist,all_avg_feat = physical_dist_priority_project(
                feat_target = target_cell_pc_feature,
                feat_source = source_cell_pc_feature,
                coords_target = coords_target,
                coords_source = coords_source,
                source_feat = source_feat,
                k2 = 1,
                pdist_thres = dist_thres,
                working_memory_t = working_memory_t)

    umap_target = sdata_inte[idx_target,:].obsm[umap_feature]
    umap_source = sdata_inte[idx_source,:].obsm[umap_feature]

    sdata_ref.layers[f'{source_sample}_raw'] = csr(all_avg_feat)
    sdata_ref.layers[f'{target_sample}_norm1e4'] = csr(sc.pp.normalize_total(sdata_ref,target_sum=1e4,layer = f'{raw_layer}',inplace=False)['X'])
    sdata_ref.layers[f'{source_sample}_norm1e4'] = csr(sc.pp.normalize_total(sdata_ref,target_sum=1e4,layer = f'{source_sample}_raw',inplace=False)['X'])
    y_true_t = np.array(sdata_inte[idx_target].obs[source_sample_ctype_col].values) if source_sample_ctype_col is not None else None
    y_source = np.array(sdata_inte[idx_source].obs[source_sample_ctype_col].values) if source_sample_ctype_col is not None else None
    y_pred_t = y_source[project_ind[:,0]] if source_sample_ctype_col is not None else None
    torch.save([physical_dist,project_ind,coords_target,coords_source,y_true_t,y_pred_t,y_source,output_path,source_sample_ctype_col,umap_target,umap_source,source_sample,target_sample,cdists,k2],f'{output_path}/mid_result{batch_t}.pt')
    if ifplot == True:
        evaluation_project(
            physical_dist = physical_dist,
            project_ind = project_ind,
            coords_target = coords_target,
            coords_source = coords_source,
            y_true_t = y_true_t,
            y_pred_t = y_pred_t,
            y_source = y_source,
            output_path = output_path,
            source_sample_ctype_col = source_sample_ctype_col,
            umap_target = umap_target,
            umap_source = umap_source,
            source_sample = source_sample,
            target_sample = target_sample,
            cdists = cdists,
            batch_t = batch_t,
            color_dict = color_dict)
    return sdata_ref,[project_ind,project_weight,cdists,physical_dist]

def average_dist(coords,quantile_t = 0.99,working_memory_t = 1000,strategy_t = 'convex'):
    coords_t = pd.DataFrame(coords)
    coords_t.drop_duplicates(inplace = True)
    coords = np.array(coords_t)
    if coords.shape[0] > 5:
        delaunay_graph_t = coords2adjacentmat(coords,output_mode='raw',strategy_t = strategy_t)
        edges = np.array(delaunay_graph_t.edges())
        def reduce_func(chunk_t, start):
            return chunk_t
        dists = pairwise_distances_chunked(coords, coords, metric='euclidean', n_jobs=-1,working_memory = working_memory_t,reduce_func = reduce_func)
        edge_dist = []
        start_t = 0
        for dist_mat_t in dists:
            end_t = start_t + dist_mat_t.shape[0]
            idx_chunk = (start_t <= edges[:,0]) & (edges[:,0] < end_t)
            edge_t = edges[idx_chunk,:]
            edge_dist_t = np.array([dist_mat_t[node - start_t,val] for [node,val] in edge_t])
            edge_dist.extend(edge_dist_t)
            start_t = end_t
        filter_thres = np.quantile(edge_dist,quantile_t)
        for i,j in edges[edge_dist > filter_thres,:]:
            delaunay_graph_t.remove_edge(i,j)
        result_t = np.mean(np.array(edge_dist)[edge_dist <= filter_thres])
        return result_t,filter_thres,edge_dist,delaunay_graph_t
    else:
        dists = pairwise_distances(coords, coords, metric='euclidean', n_jobs=-1)
        result_t = np.mean(dists.flatten())
        return result_t,'','',''

def group_shift(feat_target, feat_source, coords_target_t, coords_source_t, working_memory_t = 1000, pencentile_t = 0.8, metric_t = 'cosine'):
    from sklearn.metrics import pairwise_distances_chunked
    print(f'Using {metric_t} distance to calculate group shift:')
    feat_similarity_ctype = np.vstack(list(pairwise_distances_chunked(feat_target, feat_source, metric=metric_t, n_jobs=-1, working_memory=working_memory_t)))
    num_anchor = int(feat_similarity_ctype.shape[0] * pencentile_t)
    anchor_rank = np.argpartition(feat_similarity_ctype, num_anchor - 1, axis=-1)[:,:num_anchor]
    anchors = []
    for i in range(num_anchor):
        anchors.extend(anchor_rank[:,i].tolist())
        anchors = list(set(anchors))
        if len(anchors) >= num_anchor:
            break
    coords_shift = np.median(coords_target_t,axis=0) - np.median(coords_source_t[np.array(anchors),:],axis=0)
    return coords_shift

def physical_dist_priority_project(feat_target, feat_source, coords_target, coords_source, source_feat = None, k2 = 1, k_extend = 20, pdist_thres = 200, working_memory_t = 1000, metric_t = 'cosine'):
    def reduce_func_cdist_priority(chunk_cdist, start):
        chunk_pdist = pairwise_distances(coords_target[start:(chunk_cdist.shape[0] + start),:],coords_source, metric='euclidean', n_jobs=-1)
        idx_pdist_t = chunk_pdist < pdist_thres
        idx_pdist_sum = idx_pdist_t.sum(1)
        idx_lessk2 = (idx_pdist_sum>= k2)
        cosine_knn_ind = np.zeros([chunk_cdist.shape[0],k2]).astype(int)
        cosine_knn_weight = np.zeros_like(cosine_knn_ind).astype(float)
        cosine_knn_cdist = np.zeros_like(cosine_knn_ind).astype(float)
        cosine_knn_physical_dist = np.zeros_like(cosine_knn_ind).astype(float)
        
        idx_narrow = np.where(idx_lessk2)[0]
        idx_narrow_reverse = np.where(np.logical_not(idx_lessk2))[0]

        for i in idx_narrow:
            idx_pdist_t_i = idx_pdist_t[i,:]
            idx_i = np.where(idx_pdist_t[i,:])[0]
            knn_ind_t = idx_i[np.argpartition(chunk_cdist[i,idx_pdist_t_i], k2 - 1, axis=-1)[:k2]]
            _,weight_cell,cdist_cosine = cosine_IDW(chunk_cdist[i,knn_ind_t],k2 = k2,need_filter=False)
            cosine_knn_ind[[i],:] = knn_ind_t
            cosine_knn_weight[[i],:] = weight_cell
            cosine_knn_cdist[[i],:] = cdist_cosine
            cosine_knn_physical_dist[[i],:] = chunk_pdist[i,knn_ind_t]
        if len(idx_narrow_reverse) > 0:
            for i in idx_narrow_reverse:
                idx_pdist_extend = np.argpartition(chunk_pdist[i,:], k_extend - 1, axis=-1)[:k_extend]
                knn_ind_t = idx_pdist_extend[np.argpartition(chunk_cdist[i,idx_pdist_extend], k2 - 1, axis=-1)[:k2]]
                _,weight_cell,cdist_cosine = cosine_IDW(chunk_cdist[i,knn_ind_t],k2 = k2,need_filter=False)
                cosine_knn_ind[[i],:] = knn_ind_t
                cosine_knn_weight[[i],:] = weight_cell
                cosine_knn_cdist[[i],:] = cdist_cosine
                cosine_knn_physical_dist[[i],:] = chunk_pdist[i,knn_ind_t]
        return cosine_knn_ind,cosine_knn_weight,cosine_knn_cdist,cosine_knn_physical_dist
    
    print(f'Using {metric_t} distance to calculate cell low dimensional distance:')
    dists = pairwise_distances_chunked(feat_target, feat_source, metric=metric_t, n_jobs=-1,working_memory = working_memory_t,reduce_func=reduce_func_cdist_priority)
    cosine_knn_inds = []
    cosine_k2nn_weights = []
    cosine_k2nn_cdists = []
    cosine_k2nn_physical_dists = []
    for output in tqdm(dists):
        cosine_knn_inds.append(output[0])
        cosine_k2nn_weights.append(output[1])
        cosine_k2nn_cdists.append(output[2])
        cosine_k2nn_physical_dists.append(output[3])

    all_cosine_knn_inds = np.concatenate(cosine_knn_inds)
    all_cosine_k2nn_weights = np.concatenate(cosine_k2nn_weights)
    all_cosine_k2nn_cdists = np.concatenate(cosine_k2nn_cdists)
    all_cosine_k2nn_physical_dists = np.concatenate(cosine_k2nn_physical_dists)
    
    if source_feat is not None:
        mask_idw = sparse_mask(all_cosine_k2nn_weights,all_cosine_knn_inds, source_feat.shape[0])
        all_avg_feat = mask_idw.dot(source_feat)
        return all_cosine_knn_inds,all_cosine_k2nn_weights,all_cosine_k2nn_cdists,all_cosine_k2nn_physical_dists,all_avg_feat
    else:
        return all_cosine_knn_inds,all_cosine_k2nn_weights,all_cosine_k2nn_cdists,all_cosine_k2nn_physical_dists


def sparse_mask(idw_t, ind : np.ndarray, n_cols : int, dtype=np.float64): # ind is indices with shape (num data points, indices), in the form of output of numpy.argpartition function
    # build csr matrix from scratch
    rows = np.repeat(np.arange(ind.shape[0]), ind.shape[1]) # gives like [1,1,1,2,2,2,3,3,3]
    cols = ind.flatten() # the col indices that should be 1
    data = idw_t.flatten() # Set to `1` each (row,column) pair
    return csr_matrix((data, (rows, cols)), shape=(ind.shape[0], n_cols), dtype=dtype)

def cosine_IDW(cosine_dist_t,k2=5,eps = 1e-6,need_filter = True,ifavg = False):
    if need_filter:
        idx_cosdist_t = np.argpartition(cosine_dist_t, k2 - 1, axis=-1)[:k2]
        cdist_cosine_t = cosine_dist_t[idx_cosdist_t]
    else:
        idx_cosdist_t = 0
        cdist_cosine_t = cosine_dist_t
    if ifavg:
        weight_cell_t = np.array([1/k2] * k2)
    else:
        weight_cell_t = IDW(cdist_cosine_t,eps)
    return idx_cosdist_t, weight_cell_t, cdist_cosine_t

def IDW(df_value,eps = 1e-6):
    weights = 1.0 /(df_value + eps).T
    weights /= weights.sum(axis=0)
    return weights.T

def evaluation_project(
    physical_dist,
    project_ind,
    coords_target,
    coords_source,
    y_true_t,
    y_pred_t,
    y_source,
    output_path,
    source_sample_ctype_col,
    umap_target = None,
    umap_source = None,
    source_sample = None,
    target_sample = None,
    cdists = None,
    batch_t = '',
    exclude_group = 'Other',
    color_dict = None,
    umap_examples = False):
    print(f'Generate evaluation plots:')
    plt.rcParams.update({'pdf.fonttype':42, 'font.size' : 15})
    plt.rcParams['axes.grid'] = False
    ### histogram ###
    cdist_hist(physical_dist.flatten(),range_t = [0,2000])
    plt.savefig(f'{output_path}/physical_dist_hist{batch_t}.pdf')
    cdist_hist(cdists.flatten(),range_t = [0,2])
    plt.savefig(f'{output_path}/cdist_hist{batch_t}.pdf')

    ### confusion matrix ###
    if source_sample_ctype_col is not None:
        if exclude_group is not None:
            idx_t = y_true_t != exclude_group
            y_true_t_use = y_true_t[idx_t]
            y_pred_t_use = y_pred_t[idx_t]
        else:
            y_true_t_use = y_true_t
            y_pred_t_use = y_pred_t
        confusion_mat_plot(y_true_t_use,y_pred_t_use)
        plt.savefig(f'{output_path}/confusion_mat_raw_with_label_{source_sample_ctype_col}{batch_t}.pdf')
        confusion_mat_plot(y_true_t_use,y_pred_t_use,withlabel = False)
        plt.savefig(f'{output_path}/confusion_mat_raw_without_label_{source_sample_ctype_col}{batch_t}.pdf')

    ### link plot 3d ###
    if color_dict is not None and source_sample_ctype_col is not None:
        color_target = [color_dict[x] for x in y_true_t]
        color_source = [color_dict[x] for x in y_source]
    else:
        color_target="#9295CA" 
        color_source='#E66665'
    link_plot_3d(project_ind, coords_target, coords_source, k = 1,figsize_t = [10,10], 
                sample_n=200, link_color_mask = None, 
                color_target = color_target, color_source = color_source,
                color_true = "#222222")
    plt.savefig(f'{output_path}/link_plot{batch_t}.pdf', dpi=300)

    ### Umap ###
    if umap_examples:
        cdist_check(cdists.copy(),project_ind.copy(),umap_target,umap_source,labels_t=[target_sample,source_sample],random_seed_t=0,figsize_t=[40,32])
        plt.savefig(f'{output_path}/umap_examples{batch_t}.pdf',dpi = 300)

#################### Visualization ####################

def cdist_hist(data_t,range_t = None,step = None):
    plt.figure(figsize=[5,5])
    plt.hist(data_t, bins='auto',alpha = 0.5,color = '#1073BC')
    plt.yticks(fontsize=20)
    plt.xticks(fontsize=20)
    if type(range_t) != type(None):
        if type(step) != type(None):
            plt.xticks(np.arange(range_t[0], range_t[1] + 0.001, step),fontsize=20)
        else:
            plt.xticks(fontsize=20)
            plt.xlim(range_t[0], range_t[1])
    else:
        plt.xticks(fontsize=20)
    plt.tight_layout()

def confusion_mat_plot(y_true_t, y_pred_t, filter_thres = None, withlabel = True, fig_x = 60, fig_y = 20):
    plt.rcParams.update({'axes.labelsize' : 30,'pdf.fonttype':42,'axes.titlesize' : 30,'font.size': 15,'legend.markerscale' : 3})
    plt.rcParams['axes.grid'] = False
    TPrate = np.round(np.sum(y_pred_t == y_true_t) / len(y_true_t),2)
    uniq_t = np.unique(y_true_t,return_counts=True)
    if type(filter_thres) == type(None):
        labels_t = uniq_t[0]
    else:
        labels_t = uniq_t[0][uniq_t[1] >= filter_thres]
    plt.figure(figsize=[fig_x,fig_y])
    for idx_t, i in enumerate(['count','true','pred']):
        if i == 'count':
            normalize_t = None
            title_t = 'Counts (TP%%: %.2f)' % TPrate
        elif i == 'true':
            normalize_t = 'true'
            title_t = 'Sensitivity'
        elif i == 'pred':
            normalize_t = 'pred'
            title_t = 'Precision'
        plt.subplot(1,3,idx_t + 1)
        confusion_mat = confusion_matrix(y_true_t,y_pred_t,labels = labels_t, normalize = normalize_t)
        if i == 'count':
            vmax_t = np.max(confusion_mat)
        else:
            vmax_t = 1
        confusion_mat = pd.DataFrame(confusion_mat,columns=labels_t,index=labels_t)
        if withlabel:
            annot = np.diag(np.diag(confusion_mat.values.copy(),0),0)
            annot = np.round(annot,2)
            annot = annot.astype('str')
            annot[annot=='0.0']=''
            annot[annot=='0']=''
            sns.heatmap(confusion_mat,cmap = 'RdBu',center = 0,annot=annot,fmt='',square = True,vmax = vmax_t)
        else:
            sns.heatmap(confusion_mat,cmap = 'RdBu',center = 0,square = True,vmax = vmax_t)
        plt.title(title_t)
        plt.xlabel('Predicted label')
        plt.ylabel('True label')
        plt.tight_layout()

def cdist_check(cdist_t,cdist_idx,umap_coords0,umap_coords1, labels_t = ['query','ref'],random_seed_t = 2,figsize_t = [40,32],output_path_t = None):
    plt.rcParams.update({'xtick.labelsize' : 20,'ytick.labelsize':20, 'axes.labelsize' : 30, 'axes.titlesize' : 40,'axes.grid': False})
    random.seed(random_seed_t)
    sampled_points = np.sort(random.sample(list(range(0,cdist_idx.shape[0])),20))
    fig, axs = plt.subplots(nrows=4, ncols=5, figsize=figsize_t)
    axs = axs.flatten()
    for i in range(len(sampled_points)):
        idx_check = sampled_points[i]
        axs[i].scatter(umap_coords0[:,0],umap_coords0[:,1],s = 0.5,c = '#1f77b4',rasterized=True)
        axs[i].scatter(umap_coords1[:,0],umap_coords1[:,1],s = 0.5,c = '#E47E8B',rasterized=True)
        axs[i].scatter(umap_coords0[idx_check,0],umap_coords0[idx_check,1],s = 220,linewidth = 4,c = '#1f77b4',edgecolors = '#000000',label = labels_t[0],rasterized=False)
        axs[i].scatter(umap_coords1[cdist_idx[idx_check,0],0],umap_coords1[cdist_idx[idx_check,0],1],s = 220,linewidth=4,c = '#E47E8B',edgecolors = '#000000',label = labels_t[1], rasterized=False)
        axs[i].legend(scatterpoints=1,markerscale=2, fontsize=30)
        axs[i].set_xticks([])
        axs[i].set_yticks([])
        axs[i].set_title('cdist = ' + str(format(cdist_t[idx_check,0],'.2f')))
    if output_path_t is not None:
        plt.savefig(f'{output_path_t}/umap_examples.pdf',dpi = 300)
        plt.close('all')

def link_plot_3d(assign_mat, coords_target, coords_source, k, figsize_t = [15,20], sample_n=1000, link_color_mask=None, color_target="#9295CA", color_source='#E66665', color_true = "#999999", color_false = "#999999", remove_background = True):
    from mpl_toolkits.mplot3d.art3d import Line3DCollection
    assert k == 1
    ax = plt.figure(figsize=figsize_t).add_subplot(projection='3d')
    xylim = max(coords_source.max(), coords_target.max())
    ax.set_xlim(0, xylim)
    ax.set_ylim(0, xylim)
    ax.set_zlim(-0.1, 1.1)
    ax.set_box_aspect([1,1,0.6])
    ax.view_init(elev=25)

    coordsidx_transfer_source_link = assign_mat[:, 0]
    
    coords_transfer_source_link = coords_source[coordsidx_transfer_source_link,:]
    t1 = np.row_stack((coords_transfer_source_link[:,0],coords_transfer_source_link[:,1])) # source
    t2 = np.row_stack((coords_target[:,0],coords_target[:,1])) # target
    
    downsample_indices = np.random.choice(range(coords_target.shape[0]), sample_n)
    
    if link_color_mask is not None:
        final_true_indices = np.intersect1d(downsample_indices, np.where(link_color_mask)[0])
        final_false_indices = np.intersect1d(downsample_indices, np.where(~link_color_mask)[0])
        segs = [[(*t2[:, i], 0), (*t1[:, i], 1)] for i in final_false_indices]
        line_collection = Line3DCollection(segs, colors=color_false, lw=0.5, linestyles='dashed')
        line_collection.set_rasterized(True)
        ax.add_collection(line_collection)
    else:
        final_true_indices = downsample_indices
        
    segs = [[(*t2[:, i], 0), (*t1[:, i], 1)] for i in final_true_indices]
    line_collection = Line3DCollection(segs, colors=color_true, lw=0.5, linestyles='dashed')
    line_collection.set_rasterized(True)
    ax.add_collection(line_collection)
    
    ### target - z = 0
    ax.scatter(xs = coords_target[:,0],ys = coords_target[:,1], zs=0, s = 2, c =color_target, alpha = 0.8, ec='none', rasterized=True, depthshade=False)
    ### source - z = 1
    ax.scatter(xs = coords_source[:,0],ys = coords_source[:,1], zs=1, s = 2, c =color_source, alpha = 0.8, ec='none', rasterized=True, depthshade=False)
    if remove_background:
        # Remove axis
        ax.axis('off')
    # Remove background
    ax.xaxis.pane.fill = False
    ax.yaxis.pane.fill = False
    ax.zaxis.pane.fill = False






def add_palue(ax,line_x1,line_x2,line_y,
              text_y,text,fontsize=11,fontcolor='#000000',
             horizontalalignment='center',):
    ax.plot((line_x1,line_x2),(line_y,line_y),c=fontcolor)
    ax.text((line_x1+line_x2)/2,line_y+text_y,text,fontsize=fontsize,
            horizontalalignment=horizontalalignment,)


from __future__ import annotations

import logging
import os

import numpy as np
import pandas as pd
import scanpy as sc
import scvi

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class SCANVI_POPV(BaseAlgorithm):
    """
    Class to compute a classifier using the scANVI model.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        save_folder: str | None = None,
        result_key: str | None = "popv_scanvi_prediction",
        embedding_key: str | None = "X_scanvi_popv",
        umap_key: str | None = "X_umap_scanvi_popv",
        model_kwargs: dict | None = None,
        classifier_kwargs: dict | None = None,
        embedding_kwargs: dict | None = None,
        train_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            embedding_key=embedding_key,
        )
        self.umap_key = umap_key
        self.save_folder = save_folder

        # Initialize parameters with defaults, update with provided kwargs
        self.model_kwargs = {
            "dropout_rate": 0.05,
            "dispersion": "gene",
            "n_layers": 3,
            "n_latent": 20,
            "gene_likelihood": "nb",
            "use_batch_norm": "none",
            "use_layer_norm": "both",
            "encode_covariates": True,
        }
        if model_kwargs:
            self.model_kwargs.update(model_kwargs)

        self.train_kwargs = {
            "max_epochs": 20,
            "batch_size": 512,
            "n_samples_per_label": 20,
            "accelerator": settings.accelerator,
            "plan_kwargs": {"n_epochs_kl_warmup": 20},
            "max_epochs_unsupervised": 20,
        }
        if train_kwargs:
            self.train_kwargs.update(train_kwargs)
        # Save separate unsupervised training epoch value and remove from train_kwargs
        self.max_epochs_unsupervised = self.train_kwargs.pop("max_epochs_unsupervised")
        self.max_epochs = self.train_kwargs.get("max_epochs", None)

        self.classifier_kwargs = {"n_layers": 3, "dropout_rate": 0.1}
        if classifier_kwargs:
            self.classifier_kwargs.update(classifier_kwargs)

        self.embedding_kwargs = {"min_dist": 0.3}
        if embedding_kwargs:
            self.embedding_kwargs.update(embedding_kwargs)

    def compute_integration(self, adata):
        """
        Compute scANVI model and integrate data.
        Resulting latent representation is stored in adata.obsm[self.embedding_key].
        """
        logging.info("Integrating data with scANVI")
        if adata.uns["_prediction_mode"] == "retrain":
            if adata.uns["_pretrained_scvi_path"]:
                scvi_model = scvi.model.SCVI.load(
                    os.path.join(adata.uns["_save_path_trained_models"], "scvi"),
                    adata=adata,
                )
            else:
                scvi.model.SCVI.setup_anndata(
                    adata,
                    batch_key=self.batch_key,
                    labels_key=self.labels_key,
                    layer="scvi_counts",
                )
                scvi_model = scvi.model.SCVI(adata, **self.model_kwargs)
                scvi_model.train(
                    max_epochs=self.max_epochs_unsupervised,
                    accelerator=settings.accelerator,
                    plan_kwargs={"n_epochs_kl_warmup": 20},
                )

            self.model = scvi.model.SCANVI.from_scvi_model(
                scvi_model,
                unlabeled_category=adata.uns["unknown_celltype_label"],
                classifier_parameters=self.classifier_kwargs,
            )
        else:
            query = adata[adata.obs["_predict_cells"] == "relabel"].copy()
            self.model = scvi.model.SCANVI.load_query_data(
                query,
                os.path.join(adata.uns["_save_path_trained_models"], "scanvi"),
                freeze_classifier=True,
            )

        # Adjust training configuration for "fast" prediction mode.
        if adata.uns["_prediction_mode"] == "fast":
            self.train_kwargs["max_epochs"] = 1

        self.model.train(**self.train_kwargs)

        if adata.uns["_prediction_mode"] == "retrain":
            self.model.save(
                os.path.join(adata.uns["_save_path_trained_models"], "scanvi"),
                save_anndata=False,
                overwrite=True,
            )

        latent_representation = self.model.get_latent_representation()
        relabel_mask = adata.obs["_predict_cells"] == "relabel"
        if self.embedding_key not in adata.obsm:
            # Initialize with the correct shape if not present.
            adata.obsm[self.embedding_key] = np.zeros((adata.n_obs, latent_representation.shape[1]))
        adata.obsm[self.embedding_key][relabel_mask, :] = latent_representation

    def predict(self, adata):
        """
        Predict cell types using scANVI.
        Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving scanvi label prediction to adata.obs["{self.result_key}"]')

        # Subset once to avoid redundant computation.
        relabel_adata = adata[adata.obs["_predict_cells"] == "relabel"]

        # Predict labels once and cache the result.
        predicted_labels = self.model.predict(relabel_adata)
        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]

        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", self.result_key] = predicted_labels

        if self.return_probabilities:
            # Calculate soft predictions only once.
            soft_preds = self.model.predict(relabel_adata, soft=True)
            probs = np.max(soft_preds, axis=1)
            prob_key = f"{self.result_key}_probabilities"
            if prob_key not in adata.obs.columns:
                adata.obs[prob_key] = pd.Series(dtype="float64")
            adata.obs.loc[adata.obs["_predict_cells"] == "relabel", prob_key] = probs

    def compute_umap(self, adata):
        """
        Compute the UMAP embedding of the integrated data.
        Result is stored in adata.obsm[self.umap_key].
        """
        transformer = "rapids" if settings.cuml else None
        sc.pp.neighbors(adata, use_rep=self.embedding_key, transformer=transformer)
        method = "rapids" if settings.cuml else "umap"
        umap_result = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs)
        adata.obsm[self.umap_key] = umap_result.obsm["X_umap"]



# !/usr/bin/python
# -*- coding: utf-8 -*-
# Pytoolkit GeneSymbolUniform for hECA 2.0
# author: Yixin Chen, Haiyang Bian
# date: 2023/09/25

import scanpy as sc
import pandas as pd
import sys
import numpy as np
import scipy.sparse as sp
import os
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor
import argparse 

# functions used to construct the sparse matrix parallelly
def build_partial_csr(data_values, data_rows, data_cols, shape):
    return sp.csr_matrix((data_values, (data_rows, data_cols)), shape=shape)

def parallel_sparse_matrix(result_data_values, result_data_rows, result_data_cols, n_row, n_col, num_threads=4):
    chunk_size = len(result_data_values) // num_threads

    matrices = []
    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        futures = []
        for i in tqdm(range(num_threads), desc="Building matrices"):
            start = i * chunk_size
            end = (i + 1) * chunk_size if i != num_threads - 1 else len(result_data_values)

            futures.append(executor.submit(
                build_partial_csr, 
                result_data_values[start:end],
                result_data_rows[start:end],
                result_data_cols[start:end],
                (n_row, n_col)
            ))

        for future in tqdm(futures, desc="Collecting results"):
            matrices.append(future.result())

    return sum(matrices)

# the function to uniform gene expression matrix and save it into Anndata
def GeneSymbolUniform(input_adata=None, input_path=None, ref_table_path=None, 
                      gene_list_path=None, 
                      output_dir='./', output_prefix = "",
                      print_report=True, average_alias = True,
                      n_threads=10):
    """
    This function can be divided into following steps: 
        1) load reference table of the symbol relationships; 
        2) load the quert data (from a scanpy h5ad file with counts in 'X'); 
        3) construction the mapping dict between approved symbols and query symbols; 
        4) construct the output h5ad file and save the file and report
        
    
    :param input_adata: the input h5ad data with X as the count matrix to be uniformed.
    :type input_adata: scanpy::AnnData
    :default input_path: None
    
    :param input_path: the path of the input h5ad file (only used when input_data was not given)
    :type input_path: str
    :default input_path: None
    
    :param ref_table_path: the path of the reference table
    :type ref_table_path: str
    :default ref_table_path: "./GeneSymbolRef_upd0731.csv"
    
    :param gene_list_path: the path of the total gene list table
    :type gene_list_path: str
    :default gene_list_path: "./total_gene_list_42117.txt"
    
    :param output_dir: the path to save the output h5ad file
    :type output_dir: str
    :default output_dir: "./"
    
    :param output_prefix: the prefix of the output and report files
    :type output_prefix: str
    :default output_prefix: ""
    
    :param print_report: if True, print a report of the modified genes in the report.csv under the output_dir
    :type print_report: bool
    :defaul print_report: True
    
    :param average_alias: if average the counts of the genes mapped to the same aprroved symbol
    :type average_alias: bool
    :default average_alias: True
    
    
    :return: a h5ad data with uniformed epxression matrix.
    :rtype: scanpy::AnnData
    """

    script_dir = os.path.dirname(__file__)

    if ref_table_path is None:
        ref_table_path = os.path.join(script_dir, "GeneSymbolRef_upd0731.csv")
    if gene_list_path is None:
        gene_list_path = os.path.join(script_dir, "total_gene_list_42117.txt")

    output_dir = os.path.abspath(output_dir)
    
    log_format = "{message}"
    print(log_format)

    
    # *--------------------Load reference table--------------------*
    # Read the CSV
    ref_table_raw = pd.read_csv(ref_table_path, dtype=str,index_col=0)
    
    # *--------------------Data processing--------------------*
    # Select necessary columns
    ref_table_raw = ref_table_raw.loc[ref_table_raw.Status == 'Approved',:]
    ref_table_raw = ref_table_raw[["Approved symbol", "Alias symbol", "Previous symbol"]]

    # Filter rows with non-empty "Previous symbol" and "Alias symbol"
    ref_table_raw = ref_table_raw[(ref_table_raw["Previous symbol"].notna() & ref_table_raw["Previous symbol"] != "") |
                                  (ref_table_raw["Alias symbol"].notna() & ref_table_raw["Alias symbol"] != "")]

    # Replace "_" with "-"
    ref_table_raw["Previous symbol"] = ref_table_raw["Previous symbol"].str.replace("_", "-")
    ref_table_raw["Alias symbol"] = ref_table_raw["Alias symbol"].str.replace("_", "-")
    ref_table_raw["Approved symbol"] = ref_table_raw["Approved symbol"].str.replace("_", "-")

    # Create subsets for Previous and Alias symbols
    ref_table_prev = ref_table_raw[["Approved symbol", "Previous symbol"]].drop_duplicates()
    ref_table_prev = ref_table_prev[ref_table_prev["Previous symbol"].notna() & ref_table_prev["Previous symbol"] != ""]

    ref_table_alia = ref_table_raw[["Approved symbol", "Alias symbol"]].drop_duplicates()
    ref_table_alia = ref_table_alia[ref_table_alia["Alias symbol"].notna() & ref_table_alia["Alias symbol"] != ""]

    # *--------------------Load query data--------------------*
    if input_path is None and input_adata is None:
        # Neither of the input data form is given
        print("Error: No input data is given.")
        return(-1)
    elif not input_adata is None:
        # if adata is given, the input_path will be igored.
        adata_input = input_adata
    else:
        adata_input = sc.read_h5ad(input_path)
        
    query_data = adata_input.X
    query_gene_list = np.array(adata_input.var_names)
    print(f"The shape of query data is: {query_data.shape}")
    # logger.info("Print out first 5 genes in query data, in case something wrong happens in data loading: ")
    # logger.info(query_gene_list[:5])
    # print("Finished")

    # *--------------------Load total gene list--------------------*
    # logger.info("=========Processing Gene List=========")

    # Read the gene list from a tab-separated file
    total_gene_list_raw = pd.read_csv(gene_list_path,
                                      sep='\t', header=0, na_values=pd.NA)

    # Select the first column as the gene list
    total_gene_list = total_gene_list_raw.iloc[:, 0]

    # Replace all "_" with "-"
    total_gene_list = total_gene_list.str.replace("_", "-")
    total_gene_list = total_gene_list.values
    total_gene_list.sort()

    # Print the length of the gene list
    print(f"The length of reference gene_list is: {len(total_gene_list)}")

    # *--------------------Performing Gene Symbol Uniform--------------------*
    print("Performing gene symbol uniform, this step may take several minutes")

    # Create a DataFrame to store gene appearances
    gene_appearance_dict = {key: [] for key in total_gene_list}
    outlier_gene_list = []


    # 创建用于构建报告的列表
    report_data = []

    for query_symbol in tqdm(query_gene_list, desc="Processing"):
        report_row = {'Original.Name': query_symbol, 'Modified.Name': '', 'Status': ''}
        if query_symbol in total_gene_list:
            gene_appearance_dict[query_symbol].append(query_symbol)
            report_row['Status'] = "No Change"
        else:
            candidates_A = ref_table_alia.loc[ref_table_alia['Alias symbol'] == query_symbol, "Approved symbol"].values
            candidates_P = ref_table_prev.loc[ref_table_prev['Previous symbol'] == query_symbol, "Approved symbol"].values
            candidates = np.unique(np.hstack([candidates_A, candidates_P]))

            if len(candidates) > 1:
                report_row['Modified.Name'] = '|'.join(candidates)
                report_row['Status'] = "Multiple Candidates"
                outlier_gene_list.append(query_symbol)
            elif len(candidates) == 1:
                gene_appearance_dict[candidates[0]].append(query_symbol)
                report_row['Modified.Name'] = candidates[0]
                report_row['Status'] = "Changed"
            else:
                outlier_gene_list.append(query_symbol)
                report_row['Status'] = "Abandoned"

        report_data.append(report_row)

    # 一次性创建 DataFrame
    report = pd.DataFrame(report_data)

    # # build a DataFrame for report
    # report = pd.DataFrame(columns=['Original.Name', 'Modified.Name', 'Status'])
    # report['Original.Name'] = query_gene_list

    # for query_symbol in tqdm(query_gene_list, desc="Processing"):
    #     if query_symbol in total_gene_list:
    #         # if the query gene is in the approved list, we donot change it
    #         gene_appearance_dict[query_symbol].append(query_symbol)
    #         report.loc[report['Original.Name'] == query_symbol, 'Status'] = "No Change"
    #     else:
    #         # Matching query symbols with Alias symbol
    #         candidates_A = ref_table_alia.loc[ref_table_alia['Alias symbol'] == query_symbol,"Approved symbol"].values
    #         # Matching query symbols with Previous symbol
    #         candidates_P = ref_table_prev.loc[ref_table_prev['Previous symbol'] == query_symbol,"Approved symbol"].values

    #         candidates = np.unique(np.hstack([candidates_A,candidates_P]))

    #         if len(candidates)>1:
    #             # one input symbols with multiple aprroved symbols
    #             report.loc[report['Original.Name'] == query_symbol, 'Modified.Name'] = '|'.join(candidates)
    #             report.loc[report['Original.Name'] == query_symbol, "Status"] = "Multiple Candidates"
    #             outlier_gene_list.append(query_symbol)
    #         elif len(candidates)==1:
    #             gene_appearance_dict[candidates[0]].append(query_symbol)
    #             report.loc[report['Original.Name'] == query_symbol, 'Modified.Name'] = candidates
    #             report.loc[report['Original.Name'] == query_symbol, "Status"] = "Changed"
    #         else:
    #             # not found in the Alias symbol and Previous symbol
    #             outlier_gene_list.append(query_symbol)
    #             report.loc[report['Original.Name'] == query_symbol, "Status"] = "Abandoned"

    gene_appearance_dict_filtered =  {key: value for key, value in gene_appearance_dict.items() if len(value) > 0}


    # *--------------------Build output data--------------------*
    print("Building output data, this step may take several minutes")

    n_row = query_data.shape[0]
    n_col = len(total_gene_list)
    result_data_values = []
    result_data_rows = []
    result_data_cols = []

    total_gene_list_indices = {gene: index for index, gene in enumerate(total_gene_list)}
    query_gene_list_indices = {gene: index for index, gene in enumerate(query_gene_list)}

    query_data_csc = query_data.tocsc()  # Convert to CSC format for efficient column operations

    for key, source_genes in tqdm(gene_appearance_dict_filtered.items(), desc="Processing"):
        target_col = total_gene_list_indices.get(key)
        source_col = np.unique([query_gene_list_indices[gene] for gene in source_genes])

        col_sum = query_data_csc[:, source_col].sum(axis=1).A.flatten()  # Sum the values
        
        if average_alias:
            # this step only used when average_alias is set as True
            if len(source_col) > 1:
                col_sum = np.round(col_sum / len(source_col))

        # Add data to temporary storage
        non_zero_rows = np.where(col_sum != 0)[0]
        result_data_values.extend(col_sum[non_zero_rows])
        result_data_rows.extend(non_zero_rows)
        result_data_cols.extend([target_col] * len(non_zero_rows))

    # Construct the sparse matrix from the three lists
   
    if n_row >= 1e5:
        print(f"Because the number of cells is very large, we construct the sparse matrix parallelly with {n_threads} threads")
        result_data = parallel_sparse_matrix(result_data_values, result_data_rows, result_data_cols, n_row, n_col,n_threads)
    else:
        result_data = sp.csr_matrix((result_data_values, (result_data_rows, result_data_cols)), shape=(n_row, n_col))
    print(f"Shape of output data is {result_data.shape}. It should have 42117 genes with cell number unchanged.")
    
     # *--------------------Build output data--------------------*
    adata_output = sc.AnnData(X=result_data,obs=adata_input.obs,var=pd.DataFrame(index=total_gene_list))

    adata_output.write(os.path.join(output_dir,"{}_uniformed.h5ad".format(output_prefix)))
    print(f"h5ad file saved in:{os.path.join(output_dir,f'{output_prefix}_uniformed.h5ad')}" )
    
    if print_report:
        report.to_csv(os.path.join(output_dir,"{}_report.csv".format(output_prefix)))
        print(f"report file saved in: {os.path.join(output_dir,f'{output_prefix}_report.csv')}")
        
    return(adata_output)
    
if __name__ == "__main__" : 
    parser = argparse.ArgumentParser(description='GeneSymbolUniform parameters')
    parser.add_argument('--input_path', type=str, help='the path of the input h5ad file')
    parser.add_argument('--ref_table_path', default="./GeneSymbolRef_upd0731.csv", type=str, help='the path of the reference table.')
    parser.add_argument('--gene_list_path', default="./total_gene_list_42117.txt", type=str, help='the path of the total gene list table.')
    parser.add_argument('--output_dir', default='./', type=str, help='the path to save the output h5ad file')
    parser.add_argument('--output_prefix', default='', type=str, help='the prefix of the output file and report file')
    parser.add_argument('--print_report', action='store_true', help='print a report of the modified genes in the report.csv under the output_dir')
    parser.add_argument('--average_alias', action='store_true', help='if average the counts of the genes mapped to the same aprroved symbol')
    parser.add_argument('--n_threads',  default=1,  type=int, help='number of threads used in the construction of sparse matrix')
    args = parser.parse_args()
    GeneSymbolUniform(input_path = args.input_path, 
                      ref_table_path = args.ref_table_path, 
                      gene_list_path = args.gene_list_path, 
                      output_dir = args.output_dir, 
                      output_prefix = args.output_prefix,
                      print_report = args.print_report, 
                      average_alias = args.average_alias,
                      n_threads = args.n_threads)


import pandas as pd
import torch
import numpy as np
import anndata
import matplotlib.pyplot as plt
from ._utils import load_data, data_process,bulk2single_data_prepare
from ._vae import train_vae, generate_vae, load_vae
from ..bulk import data_drop_duplicates_index,deseq2_normalize
#from .map_utils import create_data, DFRunner, joint_analysis, knn
import os
import warnings
import matplotlib
from typing import Union,Tuple
import scanpy as sc
warnings.filterwarnings("ignore")


class Bulk2Single:
    r"""
    Bulk2Single class.
    
    """
    def __init__(self,bulk_data:pd.DataFrame,single_data:anndata.AnnData,
                 celltype_key:str,bulk_group=None,max_single_cells:int=5000,
                 top_marker_num:int=500,ratio_num:int=1,gpu:Union[int,str]=0):
        """
        Initializes the Bulk2Single class.

        Arguments:
            bulk_data: The bulk RNA-seq data.
            single_data: The single-cell RNA-seq data.
            celltype_key: The name of the column in the bulk data containing cell types.
            top_marker_num: The number of top markers to select per cell type. Default is 500.
            ratio_num: The ratio between the number of single cells and target number of converted cells. Default is 1.
            gpu: The ID of the GPU to use. Set to -1 to use CPU. Default is 0. If set to 'mps', the MPS backend will be used.

        """
        single_data.var_names_make_unique()
        bulk_data=data_drop_duplicates_index(bulk_data)
        self.bulk_data=bulk_data
        self.single_data=single_data
        
        if self.single_data.shape[0]>max_single_cells:
            print(f"......random select {max_single_cells} single cells")
            import random
            cell_idx=random.sample(self.single_data.obs.index.tolist(),max_single_cells)
            self.single_data=self.single_data[cell_idx,:]
        self.celltype_key=celltype_key
        self.bulk_group=bulk_group
        self.input_data=None
        #self.input_data=bulk2single_data_prepare(bulk_data,single_data,celltype_key)
        #self.cell_target_num = data_process(self.input_data, top_marker_num, ratio_num)

        test2=single_data.to_df()
        sc_ref=pd.DataFrame(columns=test2.columns)
        sc_ref_index=[]
        for celltype in list(set(single_data.obs[celltype_key])):
            sc_ref.loc[celltype]=single_data[single_data.obs[celltype_key]==celltype].to_df().sum()
            sc_ref_index.append(celltype)
        sc_ref.index=sc_ref_index
        self.sc_ref=sc_ref


        if gpu=='mps' and torch.backends.mps.is_available():
            print('Note that mps may loss will be nan, used it when torch is supported')
            self.used_device = torch.device("mps")
        else:
            self.used_device = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')
        self.history=[]

    def predicted_fraction(self,method='scaden',sep='\t', scaler='mms',
                        datatype='counts', genelenfile=None,
                        mode='overall', adaptive=True, variance_threshold=0.98,
                        save_model_name=None,
                        batch_size=128, epochs=128, seed=1,scale_size=2):
        from ..externel.tape import Deconvolution,ScadenDeconvolution
        sc_ref=self.sc_ref.copy()
        if method=='scaden':
            CellFractionPrediction=ScadenDeconvolution(sc_ref, 
                           self.bulk_data.T, sep=sep,
                           batch_size=batch_size, epochs=epochs)
        elif method=='tape':
            SignatureMatrix, CellFractionPrediction = \
                Deconvolution(sc_ref, self.bulk_data.T, sep=sep, scaler=scaler,
                            datatype=datatype, genelenfile=genelenfile,
                            mode=mode, adaptive=adaptive, variance_threshold=variance_threshold,
                            save_model_name=save_model_name,
                            batch_size=batch_size, epochs=epochs, seed=seed)
        else:
            raise ValueError('method must be scaden or tape')
        if self.bulk_group!=None:
            cell_total_num=self.single_data.shape[0]*self.bulk_data[self.bulk_group].mean(axis=1).sum()/self.single_data.to_df().sum().sum()
            print('Predicted Total Cell Num:',cell_total_num)
            self.cell_target_num=dict(pd.Series(CellFractionPrediction.loc[self.bulk_group].mean()*cell_total_num*scale_size).astype(int))
        
        else:
            cell_total_num=self.single_data.shape[0]*self.bulk_data.mean(axis=1).sum()/self.single_data.to_df().sum().sum()
            print('Predicted Total Cell Num:',cell_total_num)
            self.cell_target_num=dict(pd.Series(CellFractionPrediction.mean()*cell_total_num*scale_size).astype(int))
        
        return CellFractionPrediction

    def bulk_preprocess_lazy(self,)->None:
        """
        Preprocess the bulk data

        Arguments:
            group: The group of the bulk data. Default is None. It need to set to calculate the mean of each group.
        """

        print("......drop duplicates index in bulk data")
        self.bulk_data=data_drop_duplicates_index(self.bulk_data)
        print("......deseq2 normalize the bulk data")
        self.bulk_data=deseq2_normalize(self.bulk_data)
        print("......log10 the bulk data")
        self.bulk_data=np.log10(self.bulk_data+1)
        print("......calculate the mean of each group")
        if self.bulk_group is None:
            self.bulk_seq_group=self.bulk_data
            return None
        else:
            data_dg_v=self.bulk_data[self.bulk_group].mean(axis=1)
            data_dg=pd.DataFrame(index=data_dg_v.index)
            data_dg['group']=data_dg_v
            self.bulk_seq_group=data_dg
        return None
    
    def single_preprocess_lazy(self,target_sum:int=1e4)->None:
        """
        Preprocess the single data

        Arguments:
            target_sum: The target sum of the normalize. Default is 1e4.

        """

        print("......normalize the single data")
        sc.pp.normalize_total(self.single_data, target_sum=target_sum)
        print("......log1p the single data")
        sc.pp.log1p(self.single_data)
        return None
    
    def prepare_input(self,):
        print("......prepare the input of bulk2single")
        self.input_data=bulk2single_data_prepare(self.bulk_seq_group,
                                                 self.single_data,
                                                 self.celltype_key)


    def train(self,
            vae_save_dir:str='save_model',
            vae_save_name:str='vae',
            generate_save_dir:str='output',
            generate_save_name:str='output',
            batch_size:int=512,
            learning_rate:int=1e-4,
            hidden_size:int=256,
            epoch_num:int=5000,
            patience:int=50,save:bool=True)->torch.nn.Module:
        """
        Trains the VAE model.

        Arguments:
            vae_save_dir: The directory to save the trained VAE model. Default is 'save_model'.
            vae_save_name: The name of the saved VAE model. Default is 'vae'.
            generate_save_dir: The directory to save the generated single-cell data. Default is 'output'.
            generate_save_name: The name of the saved generated single-cell data. Default is 'output'.
            batch_size: The batch size for training. Default is 512.
            learning_rate: The learning rate for training. Default is 1e-4.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.
            epoch_num: The maximum number of epochs for training. Default is 5.
            patience: The number of epochs to wait before early stopping. Default is 50.
            save: Whether to save the trained VAE model. Default is True.

        Returns:
            vae_net: The trained VAE model.
        """
        if self.input_data==None:
            self.prepare_input()
        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print('...begin vae training')
        vae_net,history = train_vae(single_cell,
                            label,
                            self.used_device,
                            batch_size,
                            feature_size=feature_size,
                            epoch_num=epoch_num,
                            learning_rate=learning_rate,
                            hidden_size=hidden_size,
                            patience=patience,)
        print('...vae training done!')
        if save:
            path_save = os.path.join(vae_save_dir, f"{vae_save_name}.pth")
            if not os.path.exists(vae_save_dir):
                os.makedirs(vae_save_dir)
            torch.save(vae_net.state_dict(), path_save)
            print(f"...save trained vae in {path_save}.")
            import pickle
            #save cell_target_num
            with open(os.path.join(vae_save_dir, f"{vae_save_name}_cell_target_num.pkl"), 'wb') as f:
                pickle.dump(self.cell_target_num, f)
        self.vae_net=vae_net
        self.history=history
        return vae_net
        print('generating....')
        generate_sc_meta, generate_sc_data = generate_vae(vae_net, -1,
                                                          single_cell, label, breed_2_list,
                                                          index_2_gene, cell_number_target_num, used_device)
        sc_g=anndata.AnnData(generate_sc_data.T)
        sc_g.obs[self.celltype_key] = generate_sc_meta.loc[sc_g.obs.index,self.celltype_key].values
        sc_g.write_h5ad(os.path.join(generate_save_dir, f"{generate_save_name}.h5ad"), compression='gzip')
        self.__save_generation(generate_sc_meta, generate_sc_data,
                               generate_save_dir, generate_save_name)
        return sc_g
    
    def save(self,vae_save_dir:str='save_model',
            vae_save_name:str='vae',):
        """
        Saves the trained VAE model.

        Arguments:
            vae_save_dir: the directory to save the trained VAE model. Default is 'save_model'.
            vae_save_name: the name of the saved VAE model. Default is 'vae'.

        """
        path_save = os.path.join(vae_save_dir, f"{vae_save_name}.pth")
        if not os.path.exists(vae_save_dir):
            os.makedirs(vae_save_dir)
        torch.save(self.vae_net.state_dict(), path_save)
        import pickle
        #save cell_target_num
        with open(os.path.join(vae_save_dir, f"{vae_save_name}_cell_target_num.pkl"), 'wb') as f:
            pickle.dump(self.cell_target_num, f)
        print(f"...save trained vae in {path_save}.")
    
    def generate(self)->anndata.AnnData:
        r"""
        Generate the single-cell data.

        Returns:
            sc_g: The generated single-cell data.
        """
        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print('...generating')
        generate_sc_meta, generate_sc_data = generate_vae(self.vae_net, -1,
                                                          single_cell, label, breed_2_list,
                                                          index_2_gene, cell_number_target_num, self.used_device)
        generate_sc_meta.set_index('Cell',inplace=True)
        sc_g=anndata.AnnData(generate_sc_data.T)
        sc_g.obs[self.celltype_key] = generate_sc_meta.loc[sc_g.obs.index,'Cell_type'].values
        return sc_g
    
    def load_fraction(self,fraction_path:str):
        r"""
        Load the predicted cell fraction.

        Arguments:
            fraction_path: The path of the predicted cell fraction.

        Returns:
            fraction: The predicted cell fraction.
        """
        #load cell_target_num
        import pickle
        with open(os.path.join(fraction_path), 'rb') as f:
            self.cell_target_num = pickle.load(f)
        
    
    def load(self,vae_load_dir:str,hidden_size:int=256):
        r"""
        load the trained VAE model of Bulk2Single.

        Arguments:
            vae_load_dir: The directory to load the trained VAE model.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.
        """

        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print(f'loading model from {vae_load_dir}')
        vae_net = load_vae(feature_size, hidden_size, vae_load_dir, self.used_device)
        self.vae_net=vae_net

    def load_and_generate(self,
                              vae_load_dir:str,  # load_dir
                              hidden_size:int=256)->anndata.AnnData:
        r"""
        load the trained VAE model of Bulk2Single and generate the single-cell data.

        Arguments:
            vae_load_dir: The directory to load the trained VAE model.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.

        Returns:
            sc_g: The generated single-cell data.
        """
        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print(f'loading model from {vae_load_dir}')
        vae_net = load_vae(feature_size, hidden_size, vae_load_dir, self.used_device)
        print('...generating')
        generate_sc_meta, generate_sc_data = generate_vae(vae_net, -1,
                                                          single_cell, label, breed_2_list,
                                                          index_2_gene, cell_number_target_num, self.used_device)
        generate_sc_meta.set_index('Cell',inplace=True)
        #return generate_sc_meta, generate_sc_data
        sc_g=anndata.AnnData(generate_sc_data.T)
        sc_g.obs[self.celltype_key] = generate_sc_meta.loc[sc_g.obs.index,'Cell_type'].values
        
        print('...generating done!')
        return sc_g
    
    def filtered(self,generate_adata,highly_variable_genes:bool=True,max_value:float=10,
                     n_comps:int=100,svd_solver:str='auto',leiden_size:int=50):
        generate_adata.raw = generate_adata
        if highly_variable_genes:
            sc.pp.highly_variable_genes(generate_adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
            generate_adata = generate_adata[:, generate_adata.var.highly_variable]
        sc.pp.scale(generate_adata, max_value=max_value)
        sc.tl.pca(generate_adata, n_comps=n_comps, svd_solver=svd_solver)
        sc.pp.neighbors(generate_adata, use_rep="X_pca")
        sc.tl.leiden(generate_adata)
        filter_leiden=list(generate_adata.obs['leiden'].value_counts()[generate_adata.obs['leiden'].value_counts()<leiden_size].index)
        print("The filter leiden is ",filter_leiden)
        generate_adata=generate_adata[~generate_adata.obs['leiden'].isin(filter_leiden)]
        self.generate_adata=generate_adata.copy()

        return generate_adata
    
    def plot_loss(self,figsize:tuple=(4,4))->Tuple[matplotlib.figure.Figure,matplotlib.axes._axes.Axes]:
        r"""
        plot the loss curve of the trained VAE model.

        Arguments:
            figsize: The size of the figure. Default is (4,4).

        Returns:
            fig: The figure of the loss curve.
            ax: The axes of the figure.
        """
        fig, ax = plt.subplots(figsize=figsize)
        ax.plot(range(len(self.history)),self.history)
        ax.set_title('Beta-VAE')
        ax.set_ylabel('Loss')
        ax.set_xlabel('Epochs')
        return fig,ax


    def __get_model_input(self, data, cell_target_num):
        # input：data， celltype， bulk & output: label, dic, single_cell
        single_cell = data["input_sc_data"].values.T  # single cell data (600 * 6588)
        index_2_gene = (data["input_sc_data"].index).tolist()
        breed = data["input_sc_meta"]['Cell_type']
        breed_np = breed.values
        breed_set = set(breed_np)
        breed_2_list = list(breed_set)
        dic = {}  # breed_set to index {'B cell': 0, 'Monocyte': 1, 'Dendritic cell': 2, 'T cell': 3}
        label = []  # the label of cell (with index correspond)
        nclass = len(breed_set)

        ntrain = single_cell.shape[0]
        # FeaSize = single_cell.shape[1]
        feature_size = single_cell.shape[1]
        assert nclass == len(cell_target_num.keys()), "cell type num no match!!!"

        for i in range(len(breed_set)):
            dic[breed_2_list[i]] = i
        cell = data["input_sc_meta"]["Cell"].values

        for i in range(cell.shape[0]):
            label.append(dic[breed_np[i]])

        label = np.array(label)

        # label index the data size of corresponding target
        cell_number_target_num = {}
        for k, v in cell_target_num.items():
            cell_number_target_num[dic[k]] = v

        return single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, nclass, ntrain, feature_size

    def __save_generation(self, generate_sc_meta, generate_sc_data, generate_save_dir,
                          generate_save_name, ):
        # saving.....
        if not os.path.exists(generate_save_dir):
            os.makedirs(generate_save_dir)
        path_label_generate_csv = os.path.join(generate_save_dir, f"{generate_save_name}_sc_celltype.csv")
        path_cell_generate_csv = os.path.join(generate_save_dir, f"{generate_save_name}_sc_data.csv")

        generate_sc_meta.to_csv(path_label_generate_csv)
        generate_sc_data.to_csv(path_cell_generate_csv)
        print(f"saving to {path_label_generate_csv} and {path_cell_generate_csv}.")



import numpy as np
import scipy
import scipy.stats
import scipy.sparse 
from sklearn.decomposition import PCA,TruncatedSVD
from sklearn.neighbors import NearestNeighbors
import time

########## PREPROCESSING PIPELINE

def print_optional(string, verbose=True):
    if verbose:
        print(string)
    return

def pipeline_normalize(self, postnorm_total=None):
    ''' Total counts normalization '''
    if postnorm_total is None:
        postnorm_total = self._total_counts_obs.mean()

    self._E_obs_norm = tot_counts_norm(self._E_obs, target_total=postnorm_total, total_counts=self._total_counts_obs)

    if self._E_sim is not None:
        self._E_sim_norm = tot_counts_norm(self._E_sim, target_total=postnorm_total, total_counts=self._total_counts_sim)
    return

def pipeline_get_gene_filter(self, min_counts=3, min_cells=3, min_gene_variability_pctl=85):
    ''' Identify highly variable genes expressed above a minimum level '''
    self._gene_filter = filter_genes(self._E_obs_norm,
                                        min_counts=min_counts,
                                        min_cells=min_cells,
                                        min_vscore_pctl=min_gene_variability_pctl)
    return

def pipeline_apply_gene_filter(self):
    if self._E_obs is not None:
        self._E_obs = self._E_obs[:,self._gene_filter]
    if self._E_obs_norm is not None:
        self._E_obs_norm = self._E_obs_norm[:,self._gene_filter]
    if self._E_sim is not None:
        self._E_sim = self._E_sim[:,self._gene_filter]
    if self._E_sim_norm is not None:
        self._E_sim_norm = self._E_sim_norm[:,self._gene_filter]
    return

def pipeline_mean_center(self):
    gene_means = self._E_obs_norm.mean(0)
    self._E_obs_norm = self._E_obs_norm - gene_means
    if self._E_sim_norm is not None:
        self._E_sim_norm = self._E_sim_norm - gene_means
    return 

def pipeline_normalize_variance(self):
    gene_stdevs = np.sqrt(sparse_var(self._E_obs_norm))
    self._E_obs_norm = sparse_multiply(self._E_obs_norm.T, 1/gene_stdevs).T
    if self._E_sim_norm is not None:
        self._E_sim_norm = sparse_multiply(self._E_sim_norm.T, 1/gene_stdevs).T
    return 

def pipeline_zscore(self):
    gene_means = self._E_obs_norm.mean(0)
    gene_stdevs = np.sqrt(sparse_var(self._E_obs_norm))
    self._E_obs_norm = np.array(sparse_zscore(self._E_obs_norm, gene_means, gene_stdevs))
    if self._E_sim_norm is not None:
        self._E_sim_norm = np.array(sparse_zscore(self._E_sim_norm, gene_means, gene_stdevs))
    return

def pipeline_log_transform(self, pseudocount=1):
    self._E_obs_norm = log_normalize(self._E_obs_norm, pseudocount)
    if self._E_sim_norm is not None:
        self._E_sim_norm = log_normalize(self._E_sim_norm, pseudocount)
    return

def pipeline_truncated_svd(self, n_prin_comps=30, random_state=0, algorithm='arpack'):
    svd = TruncatedSVD(n_components=n_prin_comps, random_state=random_state, algorithm=algorithm).fit(self._E_obs_norm)
    self.set_manifold(svd.transform(self._E_obs_norm), svd.transform(self._E_sim_norm)) 
    return
    
def pipeline_pca(self, n_prin_comps=50, random_state=0, svd_solver='arpack'):
    if scipy.sparse.issparse(self._E_obs_norm):
        X_obs = self._E_obs_norm.toarray()
    else:
        X_obs = self._E_obs_norm
    if scipy.sparse.issparse(self._E_sim_norm):
        X_sim = self._E_sim_norm.toarray()
    else:
        X_sim = self._E_sim_norm

    pca = PCA(n_components=n_prin_comps, random_state=random_state, svd_solver=svd_solver).fit(X_obs)
    self.set_manifold(pca.transform(X_obs), pca.transform(X_sim)) 
    return

def matrix_multiply(X, Y):
    if not type(X) == np.ndarray:
        if scipy.sparse.issparse(X):
            X = X.toarray()
        else:
            X = np.array(X)
    if not type(Y) == np.ndarray:
        if scipy.sparse.issparse(Y):
            Y = Y.toarray()
        else:
            Y = np.array(Y)
    return np.dot(X,Y)

def log_normalize(X,pseudocount=1):
    X.data = np.log10(X.data + pseudocount)
    return X

########## LOADING DATA
def load_genes(filename, delimiter='\t', column=0, skip_rows=0):
    gene_list = []
    gene_dict = {}

    with open(filename) as f:
        for iL in range(skip_rows):
            f.readline()
        for l in f:
            gene = l.strip('\n').split(delimiter)[column]
            if gene in gene_dict:
                gene_dict[gene] += 1
                gene_list.append(gene + '__' + str(gene_dict[gene]))
                if gene_dict[gene] == 2:
                    i = gene_list.index(gene)
                    gene_list[i] = gene + '__1'
            else: 
                gene_dict[gene] = 1
                gene_list.append(gene)
    return gene_list


def make_genes_unique(orig_gene_list):
    gene_list = []
    gene_dict = {}

    for gene in orig_gene_list:
        if gene in gene_dict:
            gene_dict[gene] += 1
            gene_list.append(gene + '__' + str(gene_dict[gene]))
            if gene_dict[gene] == 2:
                i = gene_list.index(gene)
                gene_list[i] = gene + '__1'
        else:
           gene_dict[gene] = 1
           gene_list.append(gene)
    return gene_list

########## USEFUL SPARSE FUNCTIONS

def sparse_var(E, axis=0):
    ''' variance across the specified axis '''

    mean_gene = E.mean(axis=axis).A.squeeze()
    tmp = E.copy()
    tmp.data **= 2
    return tmp.mean(axis=axis).A.squeeze() - mean_gene ** 2

def sparse_multiply(E, a):
    ''' multiply each row of E by a scalar '''

    nrow = E.shape[0]
    w = scipy.sparse.lil_matrix((nrow, nrow))
    w.setdiag(a)
    return w * E

def sparse_zscore(E, gene_mean=None, gene_stdev=None):
    ''' z-score normalize each column of E '''

    if gene_mean is None:
        gene_mean = E.mean(0)
    if gene_stdev is None:
        gene_stdev = np.sqrt(sparse_var(E))
    return sparse_multiply((E - gene_mean).T, 1/gene_stdev).T

def subsample_counts(E, rate, original_totals, random_seed=0):
    if rate < 1:
        np.random.seed(random_seed)
        E.data = np.random.binomial(np.round(E.data).astype(int), rate)
        current_totals = E.sum(1).A.squeeze()
        unsampled_orig_totals = original_totals - current_totals
        unsampled_downsamp_totals = np.random.binomial(np.round(unsampled_orig_totals).astype(int), rate)
        final_downsamp_totals = current_totals + unsampled_downsamp_totals
    else:
        final_downsamp_totals = original_totals
    return E, final_downsamp_totals


########## GENE FILTERING

def runningquantile(x, y, p, nBins):

    ind = np.argsort(x)
    x = x[ind]
    y = y[ind]

    dx = (x[-1] - x[0]) / nBins
    xOut = np.linspace(x[0]+dx/2, x[-1]-dx/2, nBins)

    yOut = np.zeros(xOut.shape)

    for i in range(len(xOut)):
        ind = np.nonzero((x >= xOut[i]-dx/2) & (x < xOut[i]+dx/2))[0]
        if len(ind) > 0:
            yOut[i] = np.percentile(y[ind], p)
        else:
            if i > 0:
                yOut[i] = yOut[i-1]
            else:
                yOut[i] = np.nan

    return xOut, yOut


def get_vscores(E, min_mean=0, nBins=50, fit_percentile=0.1, error_wt=1):
    '''
    Calculate v-score (above-Poisson noise statistic) for genes in the input counts matrix
    Return v-scores and other stats
    '''

    ncell = E.shape[0]

    mu_gene = E.mean(axis=0).A.squeeze()
    gene_ix = np.nonzero(mu_gene > min_mean)[0]
    mu_gene = mu_gene[gene_ix]

    tmp = E[:,gene_ix]
    tmp.data **= 2
    var_gene = tmp.mean(axis=0).A.squeeze() - mu_gene ** 2
    del tmp
    FF_gene = var_gene / mu_gene

    data_x = np.log(mu_gene)
    data_y = np.log(FF_gene / mu_gene)

    x, y = runningquantile(data_x, data_y, fit_percentile, nBins)
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]

    gLog = lambda input: np.log(input[1] * np.exp(-input[0]) + input[2])
    h,b = np.histogram(np.log(FF_gene[mu_gene>0]), bins=200)
    b = b[:-1] + np.diff(b)/2
    max_ix = np.argmax(h)
    c = np.max((np.exp(b[max_ix]), 1))
    errFun = lambda b2: np.sum(abs(gLog([x,c,b2])-y) ** error_wt)
    b0 = 0.1
    b = scipy.optimize.fmin(func = errFun, x0=[b0], disp=False)
    a = c / (1+b) - 1


    v_scores = FF_gene / ((1+a)*(1+b) + b * mu_gene);
    CV_eff = np.sqrt((1+a)*(1+b) - 1);
    CV_input = np.sqrt(b);

    return v_scores, CV_eff, CV_input, gene_ix, mu_gene, FF_gene, a, b

def filter_genes(E, base_ix = [], min_vscore_pctl = 85, min_counts = 3, min_cells = 3, show_vscore_plot = False, sample_name = ''):
    ''' 
    Filter genes by expression level and variability
    Return list of filtered gene indices
    '''

    if len(base_ix) == 0:
        base_ix = np.arange(E.shape[0])

    Vscores, CV_eff, CV_input, gene_ix, mu_gene, FF_gene, a, b = get_vscores(E[base_ix, :])
    ix2 = Vscores>0
    Vscores = Vscores[ix2]
    gene_ix = gene_ix[ix2]
    mu_gene = mu_gene[ix2]
    FF_gene = FF_gene[ix2]
    min_vscore = np.percentile(Vscores, min_vscore_pctl)
    ix = (((E[:,gene_ix] >= min_counts).sum(0).A.squeeze() >= min_cells) & (Vscores >= min_vscore))
    
    if show_vscore_plot:
        import matplotlib.pyplot as plt
        x_min = 0.5*np.min(mu_gene)
        x_max = 2*np.max(mu_gene)
        xTh = x_min * np.exp(np.log(x_max/x_min)*np.linspace(0,1,100))
        yTh = (1 + a)*(1+b) + b * xTh
        plt.figure(figsize=(8, 6));
        plt.scatter(np.log10(mu_gene), np.log10(FF_gene), c = [.8,.8,.8], alpha = 0.3, edgecolors=None);
        plt.scatter(np.log10(mu_gene)[ix], np.log10(FF_gene)[ix], c = [0,0,0], alpha = 0.3, edgecolors=None);
        plt.plot(np.log10(xTh),np.log10(yTh));
        plt.title(sample_name)
        plt.xlabel('log10(mean)');
        plt.ylabel('log10(Fano factor)');
        plt.show()

    return gene_ix[ix]

########## CELL NORMALIZATION

def tot_counts_norm(E, total_counts = None, exclude_dominant_frac = 1, included = [], target_total = None):
    ''' 
    Cell-level total counts normalization of input counts matrix, excluding overly abundant genes if desired.
    Return normalized counts, average total counts, and (if exclude_dominant_frac < 1) list of genes used to calculate total counts 
    '''

    E = E.tocsc()
    ncell = E.shape[0]
    if total_counts is None:
        if len(included) == 0:
            if exclude_dominant_frac == 1:
                tots_use = E.sum(axis=1)
            else:
                tots = E.sum(axis=1)
                wtmp = scipy.sparse.lil_matrix((ncell, ncell))
                wtmp.setdiag(1. / tots)
                included = np.asarray(~(((wtmp * E) > exclude_dominant_frac).sum(axis=0) > 0))[0,:]
                tots_use = E[:,included].sum(axis = 1)
                print('Excluded %i genes from normalization' %(np.sum(~included)))
        else:
            tots_use = E[:,included].sum(axis = 1)
    else:
        tots_use = total_counts.copy()

    if target_total is None:
        target_total = np.mean(tots_use)

    w = scipy.sparse.lil_matrix((ncell, ncell))
    w.setdiag(float(target_total) / tots_use)
    Enorm = w * E

    return Enorm.tocsc()

########## DIMENSIONALITY REDUCTION

def get_pca(E, base_ix=[], numpc=50, keep_sparse=False, normalize=True, random_state=0, svd_solver='arpack'):
    '''
    Run PCA on the counts matrix E, gene-level normalizing if desired
    Return PCA coordinates
    '''
    # If keep_sparse is True, gene-level normalization maintains sparsity
    #     (no centering) and TruncatedSVD is used instead of normal PCA.

    if len(base_ix) == 0:
        base_ix = np.arange(E.shape[0])

    if keep_sparse:
        if normalize:
            zstd = np.sqrt(sparse_var(E[base_ix, :]))
            Z = sparse_multiply(E.T, 1 / zstd).T
        else:
            Z = E
        pca = TruncatedSVD(n_components=numpc, random_state=random_state, algorithm=svd_solver)

    else:
        if normalize:
            zmean = E[base_ix,:].mean(0)
            zstd = np.sqrt(sparse_var(E[base_ix, :]))
            Z = sparse_multiply((E - zmean).T, 1 / zstd).T
        else:
            Z = E
        pca = PCA(n_components=numpc, random_state=random_state, svd_solver=svd_solver)

    pca.fit(Z[base_ix,:])
    return pca.transform(Z)


def preprocess_and_pca(E, total_counts_normalize=True, norm_exclude_abundant_gene_frac=1, min_counts=3, min_cells=5, min_vscore_pctl=85, gene_filter=None, num_pc=50, sparse_pca=False, zscore_normalize=True, show_vscore_plot=False):
    '''
    Total counts normalize, filter genes, run PCA
    Return PCA coordinates and filtered gene indices
    '''

    if total_counts_normalize:
        print('Total count normalizing')
        E = tot_counts_norm(E, exclude_dominant_frac=norm_exclude_abundant_gene_frac)[0]

    if gene_filter is None:
        print('Finding highly variable genes')
        gene_filter = filter_genes(E, min_vscore_pctl=min_vscore_pctl, min_counts=min_counts, min_cells=min_cells, show_vscore_plot=show_vscore_plot)

    print('Using %i genes for PCA' %len(gene_filter))
    PCdat = get_pca(E[:,gene_filter], numpc=num_pc, keep_sparse=sparse_pca, normalize=zscore_normalize)

    return PCdat, gene_filter

########## GRAPH CONSTRUCTION

def get_knn_graph(X, k=5, dist_metric='euclidean', approx=False, return_edges=True, random_seed=0):
    '''
    Build k-nearest-neighbor graph
    Return edge list and nearest neighbor matrix
    '''

    t0 = time.time()
    if approx:
        try:
            from annoy import AnnoyIndex
        except:
            approx = False
            print('Could not find library "annoy" for approx. nearest neighbor search')
    if approx:
        #print('Using approximate nearest neighbor search')

        if dist_metric == 'cosine':
            dist_metric = 'angular'
        npc = X.shape[1]
        ncell = X.shape[0]
        annoy_index = AnnoyIndex(npc, metric=dist_metric)
        annoy_index.set_seed(random_seed)

        for i in range(ncell):
            annoy_index.add_item(i, list(X[i,:]))
        annoy_index.build(10) # 10 trees

        knn = []
        for iCell in range(ncell):
            knn.append(annoy_index.get_nns_by_item(iCell, k + 1)[1:])
        knn = np.array(knn, dtype=int)

    else:
        #print('Using sklearn NearestNeighbors')

        if dist_metric == 'cosine':
            nbrs = NearestNeighbors(n_neighbors=k, metric=dist_metric, algorithm='brute').fit(X)
        else:
            nbrs = NearestNeighbors(n_neighbors=k, metric=dist_metric).fit(X)
        knn = nbrs.kneighbors(return_distance=False)

    if return_edges:
        links = set([])
        for i in range(knn.shape[0]):
            for j in knn[i,:]:
                links.add(tuple(sorted((i,j))))

        t_elapse = time.time() - t0
        #print('kNN graph built in %.3f sec' %(t_elapse))

        return links, knn
    return knn

def build_adj_mat(edges, n_nodes):
    A = scipy.sparse.lil_matrix((n_nodes, n_nodes))
    for e in edges:
        i, j = e
        A[i,j] = 1
        A[j,i] = 1
    return A.tocsc()

########## 2-D EMBEDDINGS

def get_umap(X, n_neighbors=10, min_dist=0.1, metric='euclidean', random_state=0):
    import umap
    return umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist, metric=metric, random_state=random_state).fit_transform(X) 

def get_tsne(X, angle=0.5, perplexity=30, random_state=0, verbose=False):
    from sklearn.manifold import TSNE
    return TSNE(angle=angle, perplexity=perplexity, random_state=random_state, verbose=verbose).fit_transform(X)

def get_force_layout(X, n_neighbors=5, approx_neighbors=False, n_iter=300, verbose=False):
    edges = get_knn_graph(X, k=n_neighbors, approx=approx_neighbors, return_edges=True)[0]
    return run_force_layout(edges, X.shape[0], verbose=verbose)

def run_force_layout(links, n_cells, n_iter=100, edgeWeightInfluence=1, barnesHutTheta=2, scalingRatio=1, gravity=0.05, jitterTolerance=1, verbose=False):
    from fa2 import ForceAtlas2
    import networkx as nx

    G = nx.Graph()
    G.add_nodes_from(range(n_cells))
    G.add_edges_from(list(links))

    forceatlas2 = ForceAtlas2(
                  # Behavior alternatives
                  outboundAttractionDistribution=False,  # Dissuade hubs
                  linLogMode=False,  # NOT IMPLEMENTED
                  adjustSizes=False,  # Prevent overlap (NOT IMPLEMENTED)
                  edgeWeightInfluence=edgeWeightInfluence,

                  # Performance
                  jitterTolerance=jitterTolerance,  # Tolerance
                  barnesHutOptimize=True,
                  barnesHutTheta=barnesHutTheta,
                  multiThreaded=False,  # NOT IMPLEMENTED

                  # Tuning
                  scalingRatio=scalingRatio,
                  strongGravityMode=False,
                  gravity=gravity,
                  # Log
                  verbose=verbose)

    positions = forceatlas2.forceatlas2_networkx_layout(G, pos=None, iterations=n_iter)
    positions = np.array([positions[i] for i in sorted(positions.keys())])
    return positions

########## CLUSTERING

def get_spectral_clusters(A, k):
    from sklearn.cluster import SpectralClustering
    spec = SpectralClustering(n_clusters=k, random_state = 0, affinity = 'precomputed', assign_labels = 'discretize')
    return spec.fit_predict(A)


def get_louvain_clusters(nodes, edges):
    import networkx as nx
    import community
    
    G = nx.Graph()
    G.add_nodes_from(nodes)
    G.add_edges_from(edges)
    
    return np.array(list(community.best_partition(G).values()))

########## GENE ENRICHMENT

def rank_enriched_genes(E, gene_list, cell_mask, min_counts=3, min_cells=3, verbose=False):
    gix = (E[cell_mask,:]>=min_counts).sum(0).A.squeeze() >= min_cells
    print_optional('%i cells in group' %(sum(cell_mask)), verbose)
    print_optional('Considering %i genes' %(sum(gix)), verbose)
    
    gene_list = gene_list[gix]
    
    z = sparse_zscore(E[:,gix])
    scores = z[cell_mask,:].mean(0).A.squeeze()
    o = np.argsort(-scores)
    
    return gene_list[o], scores[o]
    

########## PLOTTING STUFF

def darken_cmap(cmap, scale_factor):
    cdat = np.zeros((cmap.N, 4))
    for ii in range(cdat.shape[0]):
        curcol = cmap(ii)
        cdat[ii,0] = curcol[0] * scale_factor
        cdat[ii,1] = curcol[1] * scale_factor
        cdat[ii,2] = curcol[2] * scale_factor
        cdat[ii,3] = 1
    cmap = cmap.from_list(cmap.N, cdat)
    return cmap

def custom_cmap(rgb_list):
    import matplotlib.pyplot as plt
    rgb_list = np.array(rgb_list)
    cmap = plt.cm.Reds
    cmap = cmap.from_list(rgb_list.shape[0],rgb_list)
    return cmap

def plot_groups(x, y, groups, lim_buffer = 50, saving = False, fig_dir = './', fig_name = 'fig', res = 300, close_after = False, title_size = 12, point_size = 3, ncol = 5):
    import matplotlib.pyplot as plt

    n_col = int(ncol)
    ngroup = len(np.unique(groups))
    nrow = int(np.ceil(ngroup / float(ncol)))
    fig = plt.figure(figsize = (14, 3 * nrow))
    for ii, c in enumerate(np.unique(groups)):
        ax = plt.subplot(nrow, ncol, ii+1)
        ix = groups == c

        ax.scatter(x[~ix], y[~ix], s = point_size, c = [.8,.8,.8], edgecolors=None)
        ax.scatter(x[ix], y[ix], s = point_size, c = [0,0,0], edgecolors=None)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlim([min(x) - lim_buffer, max(x) + lim_buffer])
        ax.set_ylim([min(y) - lim_buffer, max(y) + lim_buffer])

        ax.set_title(str(c), fontsize = title_size)

    fig.tight_layout()

    if saving:
        if not os.path.exists(fig_dir):
            os.makedirs(fig_dir)
        plt.savefig(fig_dir + '/' + fig_name + '.png', dpi=res)

    if close_after:
        plt.close()
